
from preql.utils import safezip, listgen, find_duplicate, SafeDict, re_split, method
from preql.context import context

from .exceptions import Signal, InsufficientAccessLevel, ReturnSignal, pql_AttributeError
from . import pql_objects as objects
from . import pql_ast as ast
from . import sql
from .interp_common import dsp, assert_type, pyvalue_inst, evaluate, cast_to_python_string, cast_to_python_int
from .state import use_scope, get_var, get_db_target, unique_name, require_access, AccessLevels, get_access_level
from .pql_types import T, Type, Id, ITEM_NAME
from .types_impl import flatten_type, pql_repr, kernel_type
from .casts import cast
from .pql_objects import AbsInstance, projected, make_instance
from .compile_binops import compile_arith, contains, compare

class AutocompleteSuggestions(Exception):
    pass

@dsp
def cast_to_instance(x: list):
    return [cast_to_instance(i) for i in x]

@dsp
def cast_to_instance(x):
    try:
        x = x.simplify()  # just compile Name?
        inst = x.compile_to_inst()
        # inst = evaluate( x)
    except ReturnSignal:
        raise Signal.make(T.CompileError, None, f"Bad compilation of {x}")

    if isinstance(inst, ast.ParameterizedSqlCode):
        raise InsufficientAccessLevel(inst)

    if not isinstance(inst, AbsInstance):
        # TODO compile error? cast error?
        # TODO need to be able to catch this above, and provide better errors
        raise Signal.make(T.TypeError, None, f"Could not compile {pql_repr(inst.type, inst)}")

    return inst



@listgen
def _process_fields(fields):
    for f in fields:
        if isinstance(f.value, objects.Function):
            # Skip functions
            continue
        try:
            v = cast_to_instance(f.value)
        except Signal as e:
            if e.type <= T.TypeError:
                raise e.replace(message=f"Cannot use object of type '{evaluate( f.value).type}' in projection.")
            raise

        # In Preql, {=>v} creates an array. In SQL, it selects the first element.
        # Here we mitigate that disparity.
        if v.type <= T.aggregated:
            v = v.primary_key()
            t = T.json_array[v.type]
            v = make_instance(sql.MakeArray(t, v.code), t, [v])

        suggested_name = str(f.name) if f.name else guess_field_name(f.value)
        name = suggested_name.rsplit('.', 1)[-1]    # Use the last attribute as name

        yield [(f.user_defined and bool(f.name), name), v]


def _exclude_items(d, req_exclude, opt_exclude):
    if not req_exclude <= set(d):
        raise ValueError(req_exclude - set(d))

    for k, v in d.items():
        if k not in req_exclude and k not in opt_exclude:
            yield k, v


@listgen
def _expand_ellipsis(obj, fields):
    direct_names = {f.value.name for f in fields if isinstance(f.value, ast.Name)}

    for f in fields:
        assert isinstance(f, ast.NamedField)

        if isinstance(f.value, ast.Ellipsis):
            if f.name:
                msg = "Cannot use a name for ellipsis (inlining operation doesn't accept a name)"
                raise Signal.make(T.SyntaxError, f, msg)

            t = obj.type
            assert t <= T.table or t <= T.struct # some_table{ ... } or some_table{ some_struct_item {...} }

            for n in f.value.exclude:
                if isinstance(n, ast.Marker):
                    raise AutocompleteSuggestions({k:(0, v) for k, v in t.elems.items()
                                                   if k not in direct_names
                                                   and k not in f.value.exclude})

                if n in direct_names:
                    raise Signal.make(T.NameError, n, f"Field to exclude '{n}' is explicitely included in projection")

            if f.value.from_struct:
                # Inline struct
                with use_scope(obj.all_attrs()):
                    s = evaluate( f.value.from_struct)
                    if not s.type <= T.struct:
                        raise Signal.make(T.TypeError, s, f"Cannot inline objects of type {s.type}")

                    items = s.attrs
            else:
                # Ellipsis for current projection
                items = obj.all_attrs()

            try:
                remaining_items = list(_exclude_items(items, set(f.value.exclude), direct_names))
            except ValueError as e:
                fte = set(e.args[0])
                raise Signal.make(T.NameError, obj, f"Fields to exclude '{fte}' not found")

            exclude = direct_names | set(f.value.exclude)
            for name, value in remaining_items:
                assert isinstance(name, str)
                assert name not in exclude
                yield ast.NamedField(name, value, user_defined=False).set_text_ref(f.text_ref)
        else:
            yield f


@dsp
def guess_field_name(_f):
    return '_'
@dsp
def guess_field_name(f: ast.Attr):
    name = f.name
    if isinstance(name, ast.Marker):
        name = '<marker>'
    return guess_field_name(f.expr) + "." + name
@dsp
def guess_field_name(f: ast.Name):
    return str(f.name)
@dsp
def guess_field_name(f: ast.Projection):
    return guess_field_name(f.table)
@dsp
def guess_field_name(f: ast.FuncCall):
    return guess_field_name(f.func)



#
# Compilation Code
#


@method
def compile_to_inst(x: objects.Object):
    return x
@method
def compile_to_inst(node: ast.Ast):
    return node


@method
def compile_to_inst(cb: ast.CodeBlock):
    if len(cb.statements) == 1:
        return cb.statements[0].compile_to_inst()

    # TODO some statements can be evaluated at compile time
    raise Signal.make(T.CompileError, cb, "Cannot compile this code block")
@method
def compile_to_inst(i: ast.If):
    cond = cast(cast_to_instance(i.cond), T.bool)
    then = cast_to_instance(i.then)
    else_ = cast_to_instance(i.else_)
    code = sql.Case(cond.code, then.code, else_.code)
    # TODO simplify this with a better type system
    res_type = kernel_type(then.type) | kernel_type(else_.type)
    inst = make_instance(code, res_type, [cond, then, else_])
    return objects.inherit_phantom_type(inst, [cond, then, else_])



@method
def compile_to_inst(proj: ast.Projection):
    table = cast_to_instance(proj.table)

    if table is objects.EmptyList:
        return table   # Empty list projection is always an empty list.

    t = T.union[T.table, T.struct]
    if not table.type <= t:
        raise Signal.make(T.TypeError, proj, f"Cannot project objects of type {table.type}")

    fields = _expand_ellipsis(table, proj.fields)

    # Test duplicates in field names. If an automatic name is used, collision should be impossible
    dup = find_duplicate([f for f in list(proj.fields) + list(proj.agg_fields) if f.name], key=lambda f: f.name)
    if dup:
        raise Signal.make(T.TypeError, dup, f"Field '{dup.name}' was already used in this projection")

    attrs = table.all_attrs()   # TODO separate here between columns and methods? (not it's done in _process_fields)

    with use_scope({n: projected(c) for n, c in attrs.items()}):
        fields = _process_fields(fields)

    for name, f in fields:
        if not f.type <= T.union[T.primitive, T.struct, T.json, T.nulltype, T.unknown]:
            raise Signal.make(T.TypeError, proj, f"Cannot project values of type: {f.type}")

    if isinstance(table, objects.StructInstance):
        d = {n[1]:c for n, c in fields}     # Remove used_defined bool
        t = T.struct({n:f.type for n, f in d.items()})
        return objects.StructInstance(t, d)

    agg_fields = []
    if proj.agg_fields:
        with use_scope({n:objects.aggregate(c) for n, c in attrs.items()}):
            agg_fields = _process_fields(proj.agg_fields)

    all_fields = fields + agg_fields
    assert all(isinstance(inst, AbsInstance) for name_, inst in all_fields)

    #
    # Make new type (and resolve names)
    #
    field_types = [(name, inst.type) for name, inst in all_fields]
    reserved_names = {name[1] for name, _ in all_fields if name[0]}
    elems = {}
    for (user_defined, name), type_ in field_types:
        # Unvectorize for placing in the table type
        type_ = kernel_type(type_)

        # Find name without collision
        if not user_defined:
            name_ = name
            i = 1
            while name in elems or name in reserved_names:
                name = name_ + str(i)
                i += 1

        assert name not in elems
        elems[name] = type_

    # TODO inherit primary key? indexes?
    # codename = state.unique_name('proj')
    new_table_type = T.table(elems, temporary=False)    # XXX abstract=True

    # Make code
    flat_codes = [code
                  for _, inst in all_fields
                  for code in inst.flatten_code()]

    sql_fields = [
        sql.ColumnAlias.make(code, nn)
        for code, (nn, _nt) in safezip(flat_codes, flatten_type(new_table_type))
    ]

    if not sql_fields:
        raise Signal.make(T.TypeError, proj, "No column provided for projection (empty projection)")

    # Make Instance
    new_table = objects.TableInstance.make(sql.null, new_table_type, [table] + [inst for _, inst in all_fields])

    groupby = []
    limit = None
    if proj.groupby:
        if fields:
            # groupby = [new_table.get_column(n).primary_key().code for n, rc in fields]
            groupby = [sql.Primitive(T.int, str(i+1)) for i in range(len(fields))]
        else:
            limit = 1
            # Alternatively we could
            #   groupby = [sql.null]
            # But postgres doesn't support it

    code = sql.Select(new_table_type, table.code, sql_fields, group_by=groupby, limit=limit)

    # Make Instance
    return new_table.replace(code=code)

@method
def compile_to_inst(order: ast.Order):
    table = cast_to_instance(order.table)
    assert_type(table.type, T.table, order, "'order'")

    with use_scope(table.all_attrs()):
        fields = cast_to_instance(order.fields)

    for f in fields:
        if not f.type <= T.primitive:
            # TODO Support 'ordering' trait?
            raise Signal.make(T.TypeError, order, f"Arguments to 'order' must be primitive")


    code = sql.table_order(table, [c.code for c in fields])

    return objects.TableInstance.make(code, table.type, [table] + fields)

@method
def compile_to_inst(expr: ast.DescOrder):
    obj = cast_to_instance(expr.value)
    return obj.replace(code=sql.Desc(obj.code))



# @method
# def compile_to_inst(lst: list):
#     return [evaluate( e) for e in lst]


@method
def compile_to_inst(o: ast.Or):
    args = cast_to_instance(o.args)
    a, b = args
    if kernel_type(a.type) != kernel_type(b.type):
        msg = f"'or' operator requires both arguments to be of the same type, got {a.type} and {b.type}"
        raise Signal.make(T.TypeError, o, msg)
    cond = cast(a, T.bool)
    code = sql.Case(cond.code, a.code, b.code)
    return objects.make_instance(code, a.type, args)

@method
def compile_to_inst(o: ast.And):
    args = cast_to_instance(o.args)
    a, b = args
    if kernel_type(a.type) != kernel_type(b.type):
        msg = f"'and' operator requires both arguments to be of the same type, got {a.type} and {b.type}"
        raise Signal.make(T.TypeError, o, msg)
    cond = cast(a, T.bool)
    code = sql.Case(cond.code, b.code, a.code)
    return objects.make_instance(code, a.type, args)

@method
def compile_to_inst(o: ast.Not):
    expr = cast_to_instance(o.expr)
    expr_bool = cast(expr, T.bool)
    code = sql.LogicalNot(expr_bool.code)
    return objects.make_instance(code, T.bool, [expr])





@method
def compile_to_inst(cmp: ast.Compare):
    insts = evaluate( cmp.args)

    if cmp.op == 'in' or cmp.op == '!in':
        return contains(cmp.op, insts[0], insts[1])

    op = {
        '==': '=',
        '<>': '!=',
    }.get(cmp.op, cmp.op)
    return compare(op, insts[0], insts[1])

@method
def compile_to_inst(neg: ast.Neg):
    expr = cast_to_instance(neg.expr)
    assert_type(expr.type, T.number, neg, "Negation")

    return make_instance(sql.Neg(expr.code), expr.type, [expr])


@method
def compile_to_inst(arith: ast.BinOp):
    args = cast_to_instance(arith.args)
    return compile_arith(arith, *args)



@method
def compile_to_inst(x: ast.Ellipsis):
    raise Signal.make(T.SyntaxError, x, "Ellipsis not allowed here")


@method
def compile_to_inst(c: ast.Const):
    if c.type == T.nulltype:
        assert c.value is None
        return objects.null
    return pyvalue_inst(c.value, c.type)

@method
def compile_to_inst(d: ast.Dict_):
    # TODO handle duplicate key names
    elems = {k or guess_field_name(v): evaluate( v) for k, v in d.elems.items()}
    t = T.struct({k: v.type for k, v in elems.items()})
    return objects.StructInstance(t, elems)


@method
def compile_to_inst(lst: objects.PythonList):
    t = lst.type.elem
    x = [sql.Primitive(t, sql._repr(t,i)) for i in lst.items]
    name = unique_name("list_")
    table_code, subq, list_type = sql.create_list(name, x)
    inst = objects.TableInstance.make(table_code, list_type, [])
    inst.subqueries[name] = subq
    return inst


@method
def compile_to_inst(lst: ast.List_):
    if not lst.elems and tuple(lst.type.elems.values()) == (T.any,):
        # XXX a little awkward
        return objects.EmptyList

    elems = evaluate( lst.elems)

    types = {e.type for e in elems}

    if len(types) > 1:
        raise Signal.make(T.TypeError, lst, f"List members must be of the same type. Got {types}")

    elem_type ,= types

    if elem_type <= T.struct:
        rows = [sql.ValuesTuple(obj.type, obj.flatten_code()) for obj in elems]
        list_type = T.table(elems=elem_type.elems)
        name = unique_name("table_")
        table_code, subq = sql.create_table(list_type, name, rows)
    else:
        if not (elem_type <= T.union[T.primitive, T.nulltype]):
            raise Signal.make(T.TypeError, lst, "Cannot create lists of type %s" % elem_type)

        assert elem_type <= lst.type.elems[ITEM_NAME], (elem_type, lst.type)

        name = unique_name("list_")
        table_code, subq, list_type = sql.create_list(name, [e.code for e in elems])

    inst = objects.TableInstance.make(table_code, list_type, elems)
    inst.subqueries[name] = subq
    return inst


# def resolve_parameters(state: State, res: ast.ResolveParameters):
@method
def compile_to_inst(res: ast.ResolveParameters):
    # XXX use a different mechanism??

    # basically cast_to_instance(). Ideally should be an instance whenever possible
    if isinstance(res.obj, objects.Instance):
        obj = res.obj
    else:
        with use_scope(res.values):
            obj = evaluate(res.obj)

        # handle non-compilable entities (meta, etc.)
        if not isinstance(obj, objects.Instance):
            if isinstance(obj, objects.Function):
                return obj
            return res.replace(obj=obj)

    require_access(AccessLevels.WRITE_DB)

    sq2 = SafeDict()
    code = _resolve_sql_parameters(obj.code, subqueries=sq2)
    subqueries = {k: _resolve_sql_parameters(v, subqueries=sq2) for k, v in obj.subqueries.items()}

    return obj.replace(code=code, subqueries=SafeDict(subqueries).update(sq2))


def _resolve_sql_parameters(compiled_sql, wrap=False, subqueries=None):
    qb = sql.QueryBuilder(get_db_target(), False)

    # Ensure <= CompiledSQL
    compiled_sql = compiled_sql.compile(qb)

    new_code = []
    for c in compiled_sql.code:
        if isinstance(c, sql.Parameter):
            inst = evaluate( get_var(c.name))
            if inst.type != c.type:
                msg = f"Internal error: Parameter is of wrong type ({c.type} != {inst.type})"
                raise Signal.make(T.CastError, None, msg)
            new_code += inst.code.compile_wrap(qb).code
            subqueries.update(inst.subqueries)
        else:
            new_code.append(c)

    res = compiled_sql.replace(code=new_code)
    if wrap:
        res = res.wrap(qb)
    return res




@method
def compile_to_inst(rps: ast.ParameterizedSqlCode):
    sql_code = cast_to_python_string(rps.string)
    if not isinstance(sql_code, str):
        raise Signal.make(T.TypeError, rps, f"Expected string, got '{rps.string}'")

    type_ = evaluate( rps.type)
    if isinstance(type_, objects.Instance):
        type_ = type_.type
    assert isinstance(type_, Type), type_
    name = unique_name("subq_")
    if type_ <= T.table:
        self_table = objects.new_table(type_, Id(name))
    else:
        self_table = None

    instances = []
    subqueries = SafeDict()
    tokens = re_split(r"\$\w+", sql_code)
    new_code = []
    for m, t in tokens:
        if m:
            assert t[0] == '$'
            if t == '$self':
                if self_table is None:
                    raise Signal.make(T.TypeError, rps, f"$self is only available for queries that return a table")
                inst = self_table
            else:
                obj = get_var(t[1:])
                if isinstance(obj, Type) and obj <= T.table:
                    # This branch isn't strictly necessary
                    # It exists to create nicer SQL code output
                    inst = objects.new_table(obj)
                else:
                    inst = cast_to_instance(obj)

            instances.append(inst)
            new_code += _resolve_sql_parameters(inst.code, wrap=bool(new_code), subqueries=subqueries).code
            assert not subqueries
        else:
            new_code.append(t)

    # TODO validation!!
    if type_ <= T.table:
        code = sql.CompiledSQL(type_, new_code, None, True, False)

        # TODO this isn't in the tests!
        fields = [sql.Name(c, path) for path, c in flatten_type(type_)]

        subq = sql.Subquery(name, fields, code)

        inst = objects.new_table(type_, Id(name), instances)
        inst.subqueries[name] = subq
        return inst

    # Cannot inherit aggregated, because some expressions such as `sum()` cancel it out.
    # XXX So why is it okay for projected?
    #     Maybe all results should be projected?
    type_ = objects.inherit_vectorized_type(type_, instances)
    code = sql.CompiledSQL(type_, new_code, None, False, False)     # XXX is False correct?
    return make_instance(code, type_, instances)

@method
def compile_to_inst(s: ast.Slice):
    obj = cast_to_instance(s.obj)

    assert_type(obj.type, T.union[T.string, T.table], s, "Slice")

    instances = [obj]
    if s.range.start:
        start = cast_to_instance(s.range.start)
        instances += [start]
    else:
        start = pyvalue_inst(0)

    if s.range.stop:
        stop = cast_to_instance(s.range.stop)
        instances += [stop]
    else:
        stop = None

    if obj.type <= T.string:
        code = sql.StringSlice(obj.code, sql.add_one(start.code), stop and sql.add_one(stop.code))
    else:
        start_n = cast_to_python_int(start)
        stop_n = stop and cast_to_python_int(stop)
        code = sql.table_slice(obj, start_n, stop_n)

    return make_instance(code, obj.type, instances)

@method
def compile_to_inst(sel: ast.Selection):
    obj = sel.table.simplify()
    if isinstance(obj, Type):
        return _apply_type_generics(obj, sel.conds)

    table = cast_to_instance(obj)

    if table.type <= T.string:
        index ,= cast_to_instance(sel.conds)
        assert index.type <= T.int, index.type
        table = table.replace(type=T.string)    # XXX why get rid of projected here? because it's a table operation node?
        slice = ast.Slice(table,
                          ast.Range(index, ast.BinOp('+', [index, ast.Const(T.int, 1)]))
                         ).set_text_ref(sel.text_ref)
        return slice.compile_to_inst()

    assert_type(table.type, T.table, sel, "Selection")

    with use_scope({n:projected(c) for n, c in table.all_attrs().items()}):
        conds = cast_to_instance(sel.conds)

    if any(t <= T.unknown for t in table.type.elem_types):
        code = sql.unknown
    else:
        for i, c in enumerate(conds):
            if not c.type <= T.bool:
                raise Signal.make(T.TypeError, sel.conds[i], f"Selection expected boolean, got {c.type}")

        code = sql.table_selection(table, [c.code for c in conds])

    return type(table).make(code, table.type, [table] + conds)

@method
def compile_to_inst(param: ast.Parameter):
    if get_access_level() == AccessLevels.COMPILE:
        if param.type <= T.struct:
            # TODO why can't I just make an instance?
            raise InsufficientAccessLevel("Structs not supported yet")
        return make_instance(sql.Parameter(param.type, param.name), param.type, [])

    return get_var(param.name)

@method
def compile_to_inst(attr: ast.Attr):
    if isinstance(attr.name, ast.Marker):
        if attr.expr:
            inst = evaluate( attr.expr)
            attrs = {k:(0,v) for k, v in inst.all_attrs().items()}
        else:
            attrs = {}
        raise AutocompleteSuggestions(attrs)

    if not attr.expr:
        raise Signal.make(T.NotImplementedError, attr, "Implicit attribute syntax not supported")

    inst = evaluate( attr.expr)
    try:
        return evaluate( inst.get_attr(attr.name))
    except pql_AttributeError as e:
        raise Signal.make(T.AttributeError, attr, e.message)



def _apply_type_generics(gen_type, type_names):
    type_objs = evaluate( type_names)
    if not type_objs:
        raise Signal.make(T.TypeError, None, f"Generics expression expected a type, got nothing.")
    for o in type_objs:
        if not isinstance(o, Type):
            if isinstance(o.code, sql.Parameter):
                # XXX hacky test, hacky solution
                raise InsufficientAccessLevel()
            raise Signal.make(T.TypeError, None, f"Generics expression expected a type, got '{o}'.")

    if len(type_objs) > 1:
        if gen_type in (T.union,):
            return gen_type(tuple(type_objs))

        raise Signal.make(T.TypeError, None, "Union types not yet supported!")

    t ,= type_objs
    try:
        return gen_type[t]
    except TypeError:
        raise Signal.make(T.TypeError, None, f"Type {t} isn't a container!")



@method
def compile_to_inst(marker: ast.Marker):
    all_vars = context.state.get_all_vars_with_rank()   # Uses overridden version of AcState
    raise AutocompleteSuggestions(all_vars)

@method
def compile_to_inst(range: ast.Range):
    target = get_db_target()
    # TODO move to sql.py
    # Requires subqueries to be part of 'code' instead of a separate 'subqueries'?
    # But then what's the point of an instance, other than carrying methods...
    start = cast_to_python_int(range.start) if range.start else 0
    if not isinstance(start, int):
        raise Signal.make(T.TypeError, range, "Range must be between integers")

    stop = None
    if range.stop:
        stop = cast_to_python_int(range.stop)
        if not isinstance(stop, int):
            raise Signal.make(T.TypeError, range, "Range must be between integers")
    elif target in (sql.mysql, sql.bigquery):
            raise Signal.make(T.NotImplementedError, range, f"{target} doesn't support infinite series!")

    type_ = T.list[T.int]

    if target == sql.bigquery:
        code = sql.RawSql(type_, f'UNNEST(GENERATE_ARRAY({start}, {stop-1})) as item')
        return objects.TableInstance.make(code, type_, [])

    if stop is None:
        stop_str = ''
    else:
        stop_str = f" WHERE item+1<{stop}"


    name = unique_name("range")
    skip = 1
    code = f"SELECT {start} AS item UNION ALL SELECT item+{skip} FROM {name}{stop_str}"
    subq = sql.Subquery(name, [], sql.RawSql(type_, code))
    code = sql.TableName(type_, Id(name))
    return objects.TableInstance(code, type_, SafeDict({name: subq}))





from lark import Token, UnexpectedCharacters, UnexpectedToken, ParseError

from preql.loggers import ac_log
from preql.utils import bfs_all_unique, dsp
from preql.context import context

from .exceptions import Signal, ReturnSignal, pql_SyntaxError
from .compiler import AutocompleteSuggestions
from .evaluate import evaluate, resolve
from .state import ThreadState, set_var, use_scope
from . import pql_ast as ast
from . import pql_objects as objects
from .pql_types import T
from . import sql, parser


@dsp
def eval_autocomplete(x, go_inside):
    _res = evaluate(x)
    # assert isinstance(res, objects.Instance)

@dsp
def eval_autocomplete(cb: ast.Statement, go_inside):
    raise NotImplementedError(cb)


@dsp
def eval_autocomplete(t: ast.Try, go_inside):
    eval_autocomplete(t.try_, go_inside)
    catch_type = evaluate(t.catch_expr)
    scope = {t.catch_name: Signal(catch_type, [], '<unknown exception>')} if t.catch_name else {}
    with use_scope(scope):
        eval_autocomplete(t.catch_block, go_inside)

@dsp
def eval_autocomplete(a: ast.InsertRows, go_inside):
    eval_autocomplete(a.value, go_inside)
@dsp
def eval_autocomplete(a: ast.Assert, go_inside):
    eval_autocomplete(a.cond, go_inside)
@dsp
def eval_autocomplete(a: ast.Print, go_inside):
    eval_autocomplete(a.value, go_inside)

@dsp
def eval_autocomplete(x: ast.If, go_inside):
    eval_autocomplete(x.then, go_inside)
    if x.else_:
        eval_autocomplete(x.else_, go_inside)

@dsp
def eval_autocomplete(x: ast.SetValue, go_inside):
    value = evaluate( x.value)
    if isinstance(x.name, ast.Name):
        set_var(x.name.name, value)

@dsp
def eval_autocomplete(cb: ast.CodeBlock, go_inside):
    for s in cb.statements[:-1]:
        eval_autocomplete(s, False)

    for s in cb.statements[-1:]:
        eval_autocomplete(s, go_inside)

@dsp
def eval_autocomplete(td: ast.TableDefFromExpr, go_inside):
    expr = evaluate(td.expr)
    assert isinstance(td.name, str)
    set_var(td.name, expr)

@dsp
def eval_autocomplete(td: ast.TableDef, go_inside):
    t = resolve(td)
    n ,= t.options['name'].parts
    set_var(n, objects.TableInstance.make(sql.unknown, t, []))

@dsp
def eval_autocomplete(td: ast.StructDef, go_inside):
    t = resolve(td)
    set_var(t.name, t)

@dsp
def eval_autocomplete(fd: ast.FuncDef, go_inside):
    f = fd.userfunc
    assert isinstance(f, objects.UserFunction)

    try:
        if go_inside:
            with use_scope({p.name:objects.unknown for p in f.params}):
                try:
                    eval_autocomplete(f.expr, go_inside)
                except ReturnSignal:
                    pass

    finally:
        cb = ast.CodeBlock([ast.Return(objects.unknown)])
        set_var(f.name, f.replace(expr=cb))

@dsp
def eval_autocomplete(r: ast.Return, go_inside):
    # Same as _execute
    value = evaluate( r.value)
    raise ReturnSignal(value)



_closing_tokens = {
    'RSQB': ']',
    'RBRACE': '}',
    'RPAR': ')',
    '$END': '<END>',
    '_NL': '\n',
}

def _search_parser(parser):
    def expand(p):
        for choice in p.choices():
            if choice in _closing_tokens:
                t = Token(choice, _closing_tokens[choice], 1, 1, 1, 1, 2, 2)
                try:
                    new_p = p.feed_token(t)
                except ParseError:    # Illegal
                    pass
                else:
                    yield new_p

    for p in bfs_all_unique([parser], expand):
        if p.result:
            return p.result

def autocomplete_tree(parser):
    if not parser:
        return

    # No marker, no autocomplete
    if 'MARKER' not in parser.choices():
        return

    # Feed marker
    t = Token('MARKER', '<MARKER>', 1, 1, 1, 1, 2, 2)
    try:
        res = parser.feed_token(t)
    except ParseError:    # Could still fail
        return

    assert not res, res

    # Search nearest solution
    return _search_parser(parser.as_immutable())


KEYWORDS = 'table update delete new func try if else for throw catch print assert const in or and not one null false true return !in'.split()
KEYWORDS = {k:(100000, None) for k in KEYWORDS}

class AcState(ThreadState):
    def get_var(self, name):
        try:
            return super().get_var(name)
        except Signal as s:
            assert s.type <= T.NameError
            return objects.UnknownInstance()

    # def get_all_vars(self):
    #     all_vars = dict(self.get_var('__builtins__').namespace)
    #     all_vars.update( self.ns.get_all_vars() )
    #     return all_vars

    def get_all_vars_with_rank(self):
        all_vars = {k:(10000, v) for k, v in self.get_var('__builtins__').namespace.items()}
        all_vars.update( self.ns.get_all_vars_with_rank() )
        all_vars.update(KEYWORDS)
        return all_vars

    def replace(self, **kw):
        assert False

def _eval_autocomplete(ac_state, stmts):
    for stmt in stmts:
        try:
            eval_autocomplete(stmt, False)
        except Signal as e:
            ac_log.exception(e)

def autocomplete(state, code, source='<autocomplete>'):
    ac_state = AcState.clone(state)
    with context(state=ac_state):
        try:
            stmts = parser.parse_stmts(code, source, wrap_syntax_error=False)
        except UnexpectedCharacters as e:
            return {}
        except UnexpectedToken as e:
            tree = autocomplete_tree(e.interactive_parser)
            if tree:
                try:
                    stmts = parser.TreeToAst(code_ref=(code, source)).transform(tree)
                except pql_SyntaxError as e:
                    return {}

                _eval_autocomplete(ac_state, stmts[:-1])

                try:
                    eval_autocomplete(stmts[-1], True)
                except AutocompleteSuggestions as e:
                    ns ,= e.args
                    return ns
                except Signal as e:
                    ac_log.exception(e)

        else:
            _eval_autocomplete(ac_state, stmts)

    return ac_state.get_all_vars_with_rank()


from contextlib import contextmanager
from functools import wraps

from . import settings
from .core.pql_ast import pyvalue, Ast
from .core import pql_objects as objects
from .core.interpreter import Interpreter
from .core.pql_types import T
from .sql_interface import create_engine
from .utils import dsp
from .core.exceptions import Signal

from .core import display
display.install_reprs()



def clean_signal(f):
    @wraps(f)
    def inner(*args, **kwargs):
        if settings.debug:
            return f(*args, **kwargs)

        try:
            return f(*args, **kwargs)
        except Signal as e:
            raise e.clean_copy() from None  # Error from Preql
    return inner


class TablePromise:
    """Returned by Preql whenever the result is a table

    Fetching values creates queries to database engine
    """

    def __init__(self, interp, inst):
        self._interp = interp
        self._inst = inst
        self._rows = None

    @property
    def type(self):
        return self._inst.type

    def to_json(self):
        "Returns table as a list of rows, i.e. ``[{col1: value, col2: value, ...}, ...]``"
        if self._rows is None:
            self._rows = self._interp.cast_to_python(self._inst)
        assert self._rows is not None
        return self._rows

    def to_pandas(self):
        "Returns table as a Pandas dataframe (requires pandas installed)"
        from pandas import DataFrame
        return DataFrame(self)

    def __eq__(self, other):
        """Compare the table to a JSON representation of it as list of objects

        Essentially: ``return self.to_json() == other``
        """
        return self.to_json() == other

    def __len__(self):
        "Run a count query on table"
        count = self._interp.call_builtin_func('count', [self._inst])
        return self._interp.cast_to_python(count)

    def __iter__(self):
        return iter(self.to_json())

    def __getitem__(self, index):
        "Run a slice query on table"
        if isinstance(index, slice):
            offset = index.start or 0
            limit = (index.stop or len(self)) - offset
            new_inst = self._interp.call_builtin_func('limit_offset', [self._inst, pyvalue(limit), pyvalue(offset)])
            return TablePromise(self._interp, new_inst)

        # TODO different debug log level / mode
        res ,= self._interp.cast_to_python(self[index:index+1]._inst)
        return res

    def __repr__(self):
        with self._interp.setup_context():
            return display.print_to_string(display.table_repr(self._inst), 'text')



@dsp
def from_python(value: TablePromise):
    return value._inst


def _prepare_instance_for_user(interp, inst):
    if inst.type <= T.table:
        return TablePromise(interp, inst)

    return interp.localize_obj(inst)


class _Delegate:

    def __init__(self, pql, fname):
        self.fname = fname
        self.pql = pql

    @clean_signal
    def __call__(self, *args, **kw):
        pql_args = [objects.from_python(a) for a in args]
        pql_kwargs = {k:objects.from_python(v) for k,v in kw.items()}
        pql_res = self.pql._interp.call_func(self.fname, pql_args, pql_kwargs)
        return self.pql._wrap_result( pql_res )


class Preql:
    """Provides an API to run Preql code from Python

    Example:
        >>> import preql
        >>> p = preql.Preql()
        >>> p('[1, 2]{item+1}')
        [2, 3]
    """

    __name__ = "Preql"

    def __init__(self, db_uri: str='sqlite://:memory:', print_sql: bool=settings.print_sql, auto_create: bool = False):
        """Initialize a new Preql instance

        Parameters:
            db_uri (str, optional): URI of database. Defaults to using a non-persistent memory database.
            print_sql (bool, optional): Whether or not to print every SQL query that is executed (default defined in settings)
        """
        self._db_uri = db_uri
        self._print_sql = print_sql
        self._auto_create = auto_create
        self._display = display.RichDisplay()
        # self.engine.ping()

        engine = create_engine(self._db_uri, print_sql=self._print_sql, auto_create=auto_create)
        self._reset_interpreter(engine)

    def __repr__(self):
        return f'Preql({self._db_uri!r}, ...)'

    def __getstate__(self):
        return self._db_uri, self._print_sql, self._display, self._interp

    def set_output_format(self, fmt):
        if fmt == 'html':
            self._display = display.HtmlDisplay()
        else:
            self._display = display.RichDisplay()

        self._interp.state.state.display = self._display  # TODO proper api


    def _reset_interpreter(self, engine=None):
        if engine is None:
            engine = self._interp.state.db
        self._interp = Interpreter(engine, self._display)
        self._interp._py_api = self # TODO proper api

    def close(self):
        self._interp.state.db.close()

    def __getattr__(self, fname):
        var = self._interp.state.get_var(fname)

        if isinstance(var, objects.Function):
            return _Delegate(self, fname)
            # @clean_signal
            # def delegate(*args, **kw):
            #     pql_args = [objects.from_python(a) for a in args]
            #     pql_kwargs = {k:objects.from_python(v) for k,v in kw.items()}
            #     pql_res = self._interp.call_func(fname, pql_args, pql_kwargs)
            #     return self._wrap_result( pql_res )
            # return delegate
        else:
            obj = self._interp.evaluate_obj( var )
            return self._wrap_result(obj)

    def _wrap_result(self, res):
        "Wraps Preql result in a Python-friendly object"
        if isinstance(res, Ast):
            raise TypeError("Returned object cannot be converted into a Python representation")
        return _prepare_instance_for_user(self._interp, res)  # TODO session, not state

    def _run_code(self, code, source_name='<api>', args=None):
        pql_args = {name: objects.from_python(value) for name, value in (args or {}).items()}
        return self._interp.execute_code(code + "\n", source_name, pql_args)

    @clean_signal
    def __call__(self, code, **args):
        res = self._run_code(code, '<inline>', args)
        if res:
            return self._wrap_result(res)

    @clean_signal
    def load(self, filename, rel_to=None):
        """Load a Preql script

        Parameters:
            filename (str): Name of script to run
            rel_to (Optional[str]): Path to which ``filename`` is relative.
        """
        self._interp.include(filename, rel_to)

    @contextmanager
    def transaction(self):
        try:
            yield self  # TODO new instance?
        except:
            self.rollback()
            raise

        self.commit()

    def start_repl(self, *args):
        "Run the interactive prompt"
        from .repl import start_repl
        start_repl(self, *args)

    def commit(self):
        return self._interp.state.db.commit()

    def rollback(self):
        return self._interp.state.db.rollback()


    def import_pandas(self, **dfs):
        """Import pandas.DataFrame instances into SQL tables

        Example:
            >>> pql.import_pandas(a=df_a, b=df_b)
        """
        return self._interp.import_pandas(dfs)


    def load_all_tables(self):
        return self._interp.load_all_tables()

    @property
    def interp(self):
        raise Exception("Reserved")
    




#     def _functions(self):
#         return {name:f for name,f in self._interp.state.namespace.items()
#                 if isinstance(f, ast.FunctionDef)}

#     def add_many(self, table, values):
#         cols = [c.name
#                 for c in self._interp.state.namespace[table].columns.values()
#                 if not isinstance(c.type, (ast.BackRefType, ast.IdType))]
#         return self.engine.addmany(table, cols, values)

#     def add(self, table, values):
#         return self.add_many(table, [values])



import operator
from preql import settings

from .pql_types import T, dp_inst
from .interp_common import pyvalue_inst, call_builtin_func 
from . import sql
from . import pql_objects as objects
from .exceptions import Signal
from .pql_objects import make_instance, remove_phantom_type
from .state import get_var, get_db_target
from .casts import cast

## Compare
def compare(op, a, b):
    res = _compare(op, remove_phantom_type(a), remove_phantom_type(b))
    return objects.inherit_phantom_type(res, [a, b])

@dp_inst
def _compare(op, a: T.any, b: T.any):
    raise Signal.make(T.TypeError, op, f"Compare not implemented for {a.type} and {b.type}")


@dp_inst
def _compare(op, _a: T.nulltype, _b: T.nulltype):
    return pyvalue_inst(op in ('=', '<=', '>='))

@dp_inst
def _compare(_op, a: T.type, _b: T.nulltype):
    assert not a.type.maybe_null()
    return pyvalue_inst(False)
@dp_inst
def _compare(op, a: T.nulltype, b: T.type):
    return _compare(op, b, a)


primitive_or_struct = T.union[T.primitive, T.struct]

@dp_inst
def _compare(op, a: T.nulltype, b: primitive_or_struct):
    # TODO Enable this type-based optimization:
    # if not b.type.nullable:
    #     return objects.pyvalue_inst(False)
    if b.type <= T.struct:
        b = b.primary_key()
    code = sql.Compare(op, [a.code, b.code])
    return objects.Instance.make(code, T.bool, [a, b])
@dp_inst
def _compare(op, a: primitive_or_struct, b: T.nulltype):
    return _compare(op, b, a)


@dp_inst
def _compare(_op, _a: T.unknown, _b: T.object):
    return objects.UnknownInstance()
@dp_inst
def _compare(_op, _a: T.object, _b: T.unknown):
    return objects.UnknownInstance()
@dp_inst
def _compare(_op, _a: T.unknown, _b: T.unknown):
    return objects.UnknownInstance()


@dp_inst
def _prepare_to_compare(op, a, b):
    if op == '=':
        return pyvalue_inst(False)
    elif op == '!=':
        return pyvalue_inst(True)
    raise Signal.make(T.TypeError, op, f"Operator '{op}' not implemented for {a.type} and {b.type}")

@dp_inst
def _prepare_to_compare(op, a: T.number | T.bool | T.t_id, b: T.number | T.bool | T.t_id):
    pass
@dp_inst
def _prepare_to_compare(op, a: T.string, b: T.string):
    pass

@dp_inst
def _compare(op, a: T.primitive, b: T.primitive):
    if settings.optimize and isinstance(a, objects.ValueInstance) and isinstance(b, objects.ValueInstance):
        f = {
            '=': operator.eq,
            '!=': operator.ne,
            '<>': operator.ne,
            '>': operator.gt,
            '<': operator.lt,
            '>=': operator.ge,
            '<=': operator.le,
        }[op]
        try:
            return pyvalue_inst(f(a.local_value, b.local_value))
        except TypeError as e:
            raise Signal.make(T.TypeError, op, f"Operator '{op}' not implemented for {a.type} and {b.type}")

    # TODO regular equality for primitives? (not 'is')
    res = _prepare_to_compare(op, a, b)
    if res is not None:
        return res

    code = sql.Compare(op, [a.code, b.code])
    return objects.Instance.make(code, T.bool, [a, b])


@dp_inst
def _compare(op, a: T.type, b: T.type):
    if op == '<=':
        return call_builtin_func("issubclass", [a, b])
    if op != '=':
        raise Signal.make(T.NotImplementedError, op, f"Cannot compare types using: {op}")
    return pyvalue_inst(a == b)

@dp_inst
def _compare(op, a: T.primitive, b: T.row):
    return _compare(op, a, b.primary_key())

@dp_inst
def _compare(op, a: T.row, b: T.primitive):
    return _compare(op, b, a)

@dp_inst
def _compare(op, a: T.row, b: T.row):
    return _compare(op, a.primary_key(), b.primary_key())



## Contains
def contains(op, a, b):
    res = _contains(op, remove_phantom_type(a), remove_phantom_type(b))
    return objects.inherit_phantom_type(res, [a, b])


@dp_inst
def _contains(op, a: T.string, b: T.string):
    f = {
        'in': 'str_contains',
        '!in': 'str_notcontains',
    }[op]
    return call_builtin_func(f, [a, b])

@dp_inst
def _contains(op, a: T.primitive, b: T.table):
    b_list = cast(b, T.list)
    if not a.type <= b_list.type.elem:
        a = cast(a, b_list.type.elem)
        # raise Signal.make(T.TypeError, op, f"Error in contains: Mismatch between {a.type} and {b.type}")

    if op == '!in':
        op = 'not in'
    code = sql.Contains(op, [a.code, b_list.code])
    return objects.Instance.make(code, T.bool, [a, b_list])

@dp_inst
def _contains(op, a: T.any, b: T.any):
    raise Signal.make(T.TypeError, op, f"Contains not implemented for {a.type} and {b.type}")



## Arith

def compile_arith(op, a, b):
    res = _compile_arith(op, remove_phantom_type(a), remove_phantom_type(b))
    return objects.inherit_phantom_type(res, [a, b])

@dp_inst
def _compile_arith(arith, a: T.any, b: T.any):
    raise Signal.make(T.TypeError, arith.op, f"Operator '{arith.op}' not implemented for {a.type} and {b.type}")



@dp_inst
def _compile_arith(arith, a: T.table, b: T.table):
    # TODO validate types
    ops = {
        "+": 'table_concat',
        "&": 'table_intersect',
        "|": 'table_union',
        "-": 'table_subtract',
    }
    # TODO compile preql funccall?
    try:
        op = ops[arith.op]
    except KeyError:
        raise Signal.make(T.TypeError, arith.op, f"Operation '{arith.op}' not supported for tables ({a.type}, {b.type})")

    return get_var(op).func(a, b)



@dp_inst
def _compile_arith(arith, a: T.string, b: T.int):
    if arith.op != '*':
        raise Signal.make(T.TypeError, arith.op, f"Operator '{arith.op}' not supported between string and integer.")
    return call_builtin_func("repeat", [a, b])


@dp_inst
def _compile_arith(arith, a: T.number, b: T.number):
    if arith.op in ('/', '**') or a.type <= T.float or b.type <= T.float:
        res_type = T.float
    else:
        res_type = T.int

    try:
        f = {
            '+': operator.add,
            '-': operator.sub,
            '*': operator.mul,
            '/': operator.truediv,
            '/~': operator.floordiv,
            '%': operator.mod,
            '**': operator.pow,
        }[arith.op]
    except KeyError:
        raise Signal.make(T.TypeError, arith, f"Operator {arith.op} not supported between types '{a.type}' and '{b.type}'")

    if settings.optimize and isinstance(a, objects.ValueInstance) and isinstance(b, objects.ValueInstance):
        # Local folding for better performance.
        # However, acts a little different than SQL. For example, in this branch 1/0 raises ValueError,
        # while SQL returns NULL
        try:
            value = f(a.local_value, b.local_value)
        except ZeroDivisionError as e:
            raise Signal.make(T.ValueError, arith.args[-1], str(e))
        if arith.op == '**':
            value = float(value)
        return pyvalue_inst(value, res_type)

    code = sql.arith(get_db_target(), res_type, arith.op, [a.code, b.code])
    return make_instance(code, res_type, [a, b])

@dp_inst
def _compile_arith(arith, a: T.string, b: T.string):
    if arith.op == 'like':
        code = sql.BinOp('like', [a.code, b.code])
        return objects.Instance.make(code, T.bool, [a, b])

    if arith.op != '+':
        raise Signal.make(T.TypeError, arith.op, f"Operator '{arith.op}' not supported for strings.")

    if settings.optimize and isinstance(a, objects.ValueInstance) and isinstance(b, objects.ValueInstance):
        # Local folding for better performance (optional, for better performance)
        return pyvalue_inst(a.local_value + b.local_value, T.string)

    code = sql.arith(get_db_target(), T.string, arith.op, [a.code, b.code])
    return make_instance(code, T.string, [a, b])




from logging import getLogger

from preql.utils import dsp
from preql.context import context

from . import pql_ast as ast
from . import pql_objects as objects
from .exceptions import Signal
from .pql_types import Type, T


logger = getLogger('interp')

# Define common dispatch functions

@dsp
def evaluate( obj: type(NotImplemented)) -> object:
    raise NotImplementedError()

@dsp
def cast_to_python(obj: type(NotImplemented)) -> object:
    raise NotImplementedError(obj)



def assert_type(t, type_, ast_node, op, msg="%s expected an object of type %s, instead got '%s'"):
    assert isinstance(t, Type), t
    assert isinstance(type_, Type)
    if not t <= type_:
        if type_.typename == 'union':
            type_str = ' or '.join("'%s'" % elem for elem in type_.elems)
        else:
            type_str = "'%s'" % type_
        raise Signal.make(T.TypeError, ast_node, msg % (op, type_str, t))

def exclude_fields(table, fields):
    proj = ast.Projection(table, [ast.NamedField(None, ast.Ellipsis(None, exclude=list(fields) ), user_defined=False)])
    return evaluate(proj)

def call_builtin_func(name, args):
    "Call a builtin pql function"
    builtins = context.state.ns.get_var('__builtins__')
    assert isinstance(builtins, objects.Module)

    expr = ast.FuncCall(builtins.namespace[name], args)
    return evaluate( expr)



def is_global_scope(state):
    assert len(state.ns) != 0
    return len(state.ns) == 1


# def cast_to_python_primitive(state, obj):
#     res = cast_to_python(obj)
#     assert isinstance(res, (int, str, float, dict, list, type(None), datetime)), (res, type(res))
#     return res

def cast_to_python_string(obj: objects.AbsInstance):
    res = cast_to_python(obj)
    if not isinstance(res, str):
        raise Signal.make(T.TypeError, obj, f"Expected string, got '{res}'")
    return res

def cast_to_python_int(obj: objects.AbsInstance):
    res = cast_to_python(obj)
    if not isinstance(res, int):
        raise Signal.make(T.TypeError, obj, f"Expected int, got '{res}'")
    return res

    
pyvalue_inst = objects.pyvalue_inst



import operator
from pathlib import Path
import subprocess
import json

import dsnparse

from .utils import classify, dataclass
from .loggers import sql_log
from .context import context

from .core.sql import Sql, QueryBuilder, sqlite, postgres, mysql, duck, bigquery, _quote
from .core.sql_import_result import sql_result_to_python, type_from_sql
from .core.pql_types import T, Type, Object, Id
from .core.exceptions import DatabaseQueryError, Signal

@dataclass
class Const(Object):
    type: Type
    value: object

class ConnectError(Exception):
    pass


def log_sql(sql):
    for i, s in enumerate(sql.split('\n')):
        prefix = '/**/    ' if i else '/**/;;  '
        sql_log.debug(prefix+s)


class SqlInterface:
    _conn: object

    def __init__(self, print_sql=False):
        self._print_sql = print_sql

    def query(self, sql, subqueries=None, qargs=(), quiet=False):
        assert context.state
        assert isinstance(sql, Sql), sql
        sql_code = self.compile_sql(sql, subqueries, qargs)

        if self._print_sql and not quiet:
            log_sql(sql_code)

        return self._execute_sql(sql.type, sql_code)

    def compile_sql(self, sql, subqueries=None, qargs=()):
        qb = QueryBuilder(self.target)

        return sql.finalize_with_subqueries(qb, subqueries)

    def commit(self):
        self._conn.commit()

    def rollback(self):
        self._conn.rollback()

    def close(self):
        self._conn.close()

    def import_table_types(self):
        # Inefficient implementation but generic
        tables = self.list_tables()
        for table_name in tables:
            table_type = self.import_table_type(table_name)
            yield None, table_name, table_type

    def qualified_name(self, name):
        return name


# from multiprocessing import Queue
import queue
import threading
from time import sleep

class TaskQueue:
    def __init__(self):
        self._queue = queue.Queue()
        self._task_results = {}
        self._closed = False
        self._start_worker()

    def _add_task(self, task, *args, **kwargs):
        args = args or ()
        kwargs = kwargs or {}
        task_id = object()
        self._queue.put((task_id, task, args, kwargs))
        return task_id

    def _start_worker(self):
        self.worker = t = threading.Thread(target=self._worker)
        t.daemon = True
        t.start()

    def _worker(self):
        while not self._closed:
            try:
                task_id, item, args, kwargs = self._queue.get(timeout=0.1)
            except queue.Empty:
                continue
            try:
                res = item(*args, **kwargs)  
            except Exception as e:
                res = e
            self._task_results[task_id] = res
            self._queue.task_done()


    def _get_result(self, task_id):
        while task_id not in self._task_results:
            sleep(0.001)
        res = self._task_results.pop(task_id)
        if isinstance(res, Exception):
            raise res
        return res

    def run_task(self, task, *args, **kwargs):
        task_id = self._add_task(task, *args, **kwargs)
        return self._get_result(task_id)

    def close(self):
        self._closed = True



class BaseConnection:
    pass

class ThreadedConnection(BaseConnection):
    def __init__(self, create_connection):
        self._queue = TaskQueue()
        self._conn = self._queue.run_task(create_connection)

    def _backend_execute_sql(self, sql_code):
        c = self._conn.cursor()
        c.execute(sql_code)
        return c

    def _import_result(self, sql_type, c):
        if sql_type is T.nulltype:
            return None

        try:
            res = c.fetchall()
        except Exception as e:
            msg = "Exception when trying to fetch SQL result. Got error: %s"
            raise DatabaseQueryError(msg%(e))

        return sql_result_to_python(Const(sql_type, res))

    def _execute_sql(self, state, sql_type, sql_code):
        assert state
        with context(state=state):
            try:
                c = self._backend_execute_sql(sql_code)
            except Exception as e:
                msg = "Exception when trying to execute SQL code:\n    %s\n\nGot error: %s"
                raise DatabaseQueryError(msg%(sql_code, e))

            return self._import_result(sql_type, c)

    def execute_sql(self, sql_type, sql_code):
        return self._queue.run_task(self._execute_sql, context.state, sql_type, sql_code)


    def commit(self):
        self._queue.run_task(self._conn.commit)

    def rollback(self):
        self._queue.run_task(self._conn.rollback)

    def close(self):
        self._queue.run_task(self._conn.close)
        self._queue.close()




class SqlInterfaceCursor(SqlInterface):
    "An interface that uses the standard SQL cursor interface"

    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)

        self._conn = ThreadedConnection(self._create_connection)


    def _execute_sql(self, sql_type, sql_code):
        return self._conn.execute_sql(sql_type, sql_code)

    def ping(self):
        c = self._conn.cursor()
        c.execute('select 1')
        row ,= c.fetchall()
        n ,= row
        assert n == 1



class MysqlInterface(SqlInterfaceCursor):
    target = mysql

    def __init__(self, host, port, database, user, password, print_sql=False):
        self._print_sql = print_sql

        args = dict(host=host, port=port, database=database, user=user, password=password)
        self._args = {k:v for k, v in args.items() if v is not None}
        super().__init__(print_sql)

    def _create_connection(self):
        import mysql.connector
        from mysql.connector import errorcode

        try:
            return mysql.connector.connect(charset='utf8', use_unicode=True, **self._args)
        except mysql.connector.Error as e:
            if e.errno == errorcode.ER_ACCESS_DENIED_ERROR:
                raise ConnectError("Bad user name or password") from e
            elif e.errno == errorcode.ER_BAD_DB_ERROR:
                raise ConnectError("Database does not exist") from e
            else:
                raise ConnectError(*e.args) from e

    def table_exists(self, name):
        tables = [t.lower() for t in self.list_tables()]
        return name.lower() in tables

    def list_tables(self):
        sql_code = "SHOW TABLES"
        return self._execute_sql(T.list[T.string], sql_code)

    def import_table_type(self, name, columns_whitelist=None):
        columns_t = T.table(dict(
            name=T.string,
            type=T.string,
            nullable=T.string,
            key=T.string,
            default=T.string,
            extra=T.string,
        ))
        columns_q = "desc %s" % name
        sql_columns = self._execute_sql(columns_t, columns_q)

        if columns_whitelist:
            wl = set(columns_whitelist)
            sql_columns = [c for c in sql_columns if c['name'] in wl]

        cols = {c['name']: type_from_sql(c['type'].decode(), c['nullable']) for c in sql_columns}

        return T.table(cols, name=Id(name))


class PostgresInterface(SqlInterfaceCursor):
    target = postgres

    def __init__(self, host, port, database, user, password, print_sql=False):
        self.args = dict(host=host, port=port, database=database, user=user, password=password)
        super().__init__(print_sql)

    def _create_connection(self):
        import psycopg2
        import psycopg2.extras
        psycopg2.extensions.set_wait_callback(psycopg2.extras.wait_select)
        try:
            return psycopg2.connect(**self.args)
        except psycopg2.OperationalError as e:
            raise ConnectError(*e.args) from e




    def table_exists(self, name):
        sql_code = "SELECT count(*) FROM information_schema.tables where table_name='%s'" % name
        cnt = self._execute_sql(T.int, sql_code)
        return cnt > 0

    def list_tables(self):
        sql_code = "SELECT table_name FROM information_schema.tables where table_schema='public'"
        return self._execute_sql(T.list[T.string], sql_code)


    _schema_columns_t = T.table(dict(
        schema=T.string,
        table=T.string,
        name=T.string,
        pos=T.int,
        nullable=T.bool,
        type=T.string,
    ))

    def import_table_type(self, name, columns_whitelist=None):

        columns_q = """SELECT table_schema, table_name, column_name, ordinal_position, is_nullable, data_type
            FROM information_schema.columns
            WHERE table_name = '%s'
            """ % name
        sql_columns = self._execute_sql(self._schema_columns_t, columns_q)

        if columns_whitelist:
            wl = set(columns_whitelist)
            sql_columns = [c for c in sql_columns if c['name'] in wl]

        cols = [(c['pos'], c['name'], type_from_sql(c['type'], c['nullable'])) for c in sql_columns]
        cols.sort()
        cols = dict(c[1:] for c in cols)

        return T.table(cols, name=Id(name))

    def import_table_types(self):
        columns_q = """SELECT table_schema, table_name, column_name, ordinal_position, is_nullable, data_type
            FROM information_schema.columns
            """
        sql_columns = self._execute_sql(self._schema_columns_t, columns_q)

        columns_by_table = classify(sql_columns, lambda c: (c['schema'], c['table']))

        for (schema, table_name), columns in columns_by_table.items():
            cols = [(c['pos'], c['name'], type_from_sql(c['type'], c['nullable'])) for c in columns]
            cols.sort()
            cols = dict(c[1:] for c in cols)

            # name = '%s.%s' % (schema, table_name)
            yield schema, table_name, T.table(cols, name=Id(schema, table_name))



class BigQueryInterface(SqlInterface):
    target = bigquery

    def __init__(self, project, print_sql=False):
        from google.cloud import bigquery

        self._client = bigquery.Client(project)
        self._default_dataset = None

        self._print_sql = print_sql


    def table_exists(self, name):
        from google.api_core.exceptions import NotFound
        try:
            self._client.get_table(name)
        except NotFound:
            return False
        return True

    def _list_tables(self):
        for ds in self._client.list_datasets():
            for t in self._client.list_tables(ds.reference):
                yield t.full_table_id.replace(':', '.')     # Hacky

    def list_tables(self):
        return list(self._list_tables())


    def _execute_sql(self, sql_type, sql_code):
        assert context.state
        try:
            res = list(self._client.query(sql_code))
        except Exception as e:
            msg = "Exception when trying to execute SQL code:\n    %s\n\nGot error: %s"
            raise DatabaseQueryError(msg%(sql_code, e))

        if sql_type is not T.nulltype:
            res = [list(i.values()) for i in res]
            return sql_result_to_python(Const(sql_type, res))



    _schema_columns_t = T.table(dict(
        schema=T.string,
        table=T.string,
        name=T.string,
        pos=T.int,
        nullable=T.bool,
        type=T.string,
    ))

    def import_table_type(self, name, columns_whitelist=None):
        cols = {}
        for f in self._client.get_table(name).schema:
            if columns_whitelist is None or f.name in columns_whitelist:
                cols[f.name] = type_from_sql(f.field_type, f.is_nullable)

        return T.table(cols, name=Id(name))

    def list_datasets(self):
        return [ds.dataset_id for ds in self._client.list_datasets()]

    def set_default_dataset(self, dataset):
        self._default_dataset = dataset

    def get_default_dataset(self):
        if self._default_dataset is None:
            datasets = self.list_datasets()
            if not datasets:
                raise Signal(T.ValueError, None, "No dataset found.")
            self._default_dataset = datasets[0]
        return self._default_dataset

    def qualified_name(self, name):
        if '.' in name: # already has dataset
            return name
        return self.get_default_dataset() + '.' + name

    def rollback(self):
        # XXX No error? No warning?
        pass
    def commit(self):
        # XXX No error? No warning?
        pass

class AbsSqliteInterface:
    def table_exists(self, name):
        sql_code = "SELECT count(*) FROM sqlite_master where name='%s' and type='table'" % name
        cnt = self._execute_sql(T.int, sql_code)
        return cnt > 0

    def list_tables(self):
        sql_code = "SELECT name FROM sqlite_master where type='table'"
        return self._execute_sql(T.list[T.string], sql_code)

    table_schema_type = T.table(dict(
        pos=T.int,
        name=T.string,
        type=T.string,
        notnull=T.bool,
        default_value=T.string,
        pk=T.bool,
    ))

    def import_table_type(self, name, columns_whitelist=None):
        columns_q = """pragma table_info('%s')""" % name
        sql_columns = self._execute_sql(self.table_schema_type, columns_q)

        if columns_whitelist:
            wl = set(columns_whitelist)
            sql_columns = [c for c in sql_columns if c['name'] in wl]

        cols = [(c['pos'], c['name'], type_from_sql(c['type'], not c['notnull'])) for c in sql_columns]
        cols.sort()
        cols = dict(c[1:] for c in cols)

        pk = [[c['name']] for c in sql_columns if c['pk']]

        return T.table(cols, name=Id(name), pk=pk)


class _SqliteProduct:
    def __init__(self):
        self.product = 1

    def step(self, value):
        self.product *= value

    def finalize(self):
        return self.product

import math
class _SqliteStddev:
    def __init__(self):
        self.M = 0.0
        self.S = 0.0
        self.k = 1

    def step(self, value):
        if value is None:
            return
        tM = self.M
        self.M += (value - tM) / self.k
        self.S += (value - tM) * (value - self.M)
        self.k += 1

    def finalize(self):
        if self.k < 3:
            return None
        return math.sqrt(self.S / (self.k-2))



class SqliteInterface(SqlInterfaceCursor, AbsSqliteInterface):
    target = sqlite

    def __init__(self, filename=None, print_sql=False):
        self._filename = filename
        super().__init__(print_sql)


    def _create_connection(self):
        import sqlite3
        # sqlite3.enable_callback_tracebacks(True)
        try:
            conn = sqlite3.connect(self._filename or ':memory:')
        except sqlite3.OperationalError as e:
            raise ConnectError(*e.args) from e

        def sqlite_throw(x):
            raise Exception(x)

        conn.create_function("power", 2, operator.pow)
        conn.create_function("_pql_throw", 1, sqlite_throw)
        conn.create_aggregate("_pql_product", 1, _SqliteProduct)
        conn.create_aggregate("stddev", 1, _SqliteStddev)
        return conn


class DuckInterface(AbsSqliteInterface):
    target = duck

    def __init__(self, filename=None, print_sql=False):
        import duckdb
        self._conn = duckdb.connect(filename or ':memory:')
        self._print_sql = print_sql

    def rollback(self):
        pass    # XXX


class GitInterface(AbsSqliteInterface):
    "Uses https://github.com/augmentable-dev/askgit"

    target = sqlite

    def __init__(self, path, print_sql):
        self.path = path
        self._print_sql = print_sql


    table_schema_type = T.table(dict(
        cid=T.int,
        dflt_value=T.string,
        name=T.string,
        notnull=T.bool,
        pk=T.bool,
        type=T.string,
    ))

    def _execute_sql(self, sql_type, sql_code):
        assert context.state
        try:
            res = subprocess.check_output(['askgit', '--format', 'json', sql_code])
        except FileNotFoundError:
            msg = "Could not find executable 'askgit'. Make sure it's installed, and try again."
            raise DatabaseQueryError(msg)
        except subprocess.CalledProcessError as e:
            msg = "Exception when trying to execute SQL code:\n    %s\n\nGot error: %s"
            raise DatabaseQueryError(msg%(sql_code, e))

        return self._import_result(sql_type, res)

    def _import_result(self, sql_type, c):
        if sql_type is not T.nulltype:
            if sql_type <= T.table:
                lookup = dict(reversed(x) for x in enumerate(sql_type.elems))
                rows = [json.loads(x) for x in c.split(b'\n') if x.strip()]
                res = []
                for row in rows:
                    # TODO refactor into a function
                    x = ["PLACEHOLDER"] * len(row)
                    for k, v in row.items():
                        x[lookup[k]] = v
                    assert "PLACEHOLDER" not in x, (x, row)
                    res.append(x)
            else:
                res = [list(json.loads(x).values()) for x in c.split(b'\n') if x.strip()]

            return sql_result_to_python(Const(sql_type, res))

    def import_table_type(self, name, columns_whitelist=None):
        # TODO merge with superclass

        columns_q = """pragma table_info('%s')""" % name
        sql_columns = self._execute_sql(self.table_schema_type, columns_q)

        if columns_whitelist:
            wl = set(columns_whitelist)
            sql_columns = [c for c in sql_columns if c['name'] in wl]

        cols = [(c['cid'], c['name'], type_from_sql(c['type'], not c['notnull'])) for c in sql_columns]
        cols.sort()
        cols = dict(c[1:] for c in cols)

        return T.table(cols, name=Id(name))

    def list_tables(self):
        # TODO merge with superclass?
        sql_code = "SELECT name FROM sqlite_master where type='table'"
        res = self._execute_sql(T.table(dict(name=T.string)), sql_code)
        return [x['name'] for x in res]




def _drop_tables(state, *tables):
    # XXX temporary. Used for testing
    db = state.db
    with context(state=state):
        for t in tables:
            t = _quote(db.target, db.qualified_name(t))
            db._execute_sql(T.nulltype, f"DROP TABLE {t};")


_SQLITE_SCHEME = 'sqlite://'

def create_engine(db_uri, print_sql, auto_create):
    if db_uri.startswith(_SQLITE_SCHEME):
        # Parse sqlite:// ourselves, to allow for sqlite://c:/path/to/db
        path = db_uri[len(_SQLITE_SCHEME):]
        if not auto_create and path != ':memory:':
            if not Path(path).exists():
                raise ConnectError("File %r doesn't exist. To create it, set auto_create to True" % path)
        return SqliteInterface(path, print_sql=print_sql)

    dsn = dsnparse.parse(db_uri)
    if len(dsn.schemes) > 1:
        raise NotImplementedError("Preql doesn't support multiple schemes")
    scheme ,= dsn.schemes

    if len(dsn.paths) != 1:
        raise ValueError("Bad value for uri: %s" % db_uri)
    path ,= dsn.paths

    if scheme == 'postgres':
        return PostgresInterface(dsn.host, dsn.port, path, dsn.user, dsn.password, print_sql=print_sql)
    elif scheme == 'mysql':
        return MysqlInterface(dsn.host, dsn.port, path, dsn.user, dsn.password, print_sql=print_sql)
    elif scheme == 'git':
        return GitInterface(path, print_sql=print_sql)
    elif scheme == 'duck':
        return DuckInterface(path, print_sql=print_sql)
    elif scheme == 'bigquery':
        return BigQueryInterface(path, print_sql=print_sql)

    raise NotImplementedError(f"Scheme {dsn.scheme} currently not supported")



from .exceptions import Signal, pql_AttributeError

from preql.utils import concat_for, classify_bool

from .base import Object
from .pql_types import ITEM_NAME, T, Type, dp_type


def Object_get_attr(self, attr):
    raise pql_AttributeError(attr)

def Object_isa(self, t):
    if not isinstance(t, Type):
        raise Signal.make(T.TypeError, None, f"'type' argument to isa() isn't a type. It is {t}")
    return self.type <= t

Object.get_attr = Object_get_attr
Object.isa = Object_isa

def _type_flatten_code(self):
    raise Signal.make(T.TypeError, None, f"Found type 'type' in unexpected place")

Type.flatten_code = _type_flatten_code

@dp_type
def flatten_path(path, t):
    return [(path, t)]

@dp_type
def flatten_path(path, t: T.union[T.table, T.struct]):
    elems = t.elems
    if t.maybe_null():
        elems = {k:v.as_nullable() for k, v in elems.items()}
    return concat_for(flatten_path(path + [name], col) for name, col in elems.items())
@dp_type
def flatten_path(path, t: T.list):
    return concat_for(flatten_path(path + [name], col) for name, col in [(ITEM_NAME, t.elem)])


def flatten_type(tp, path = []):
    return [(join_names(path), t) for path, t in flatten_path(path, tp)]


def table_params(t):
    # TODO hide_from_init / writeable, not id
    return [(name, c) for name, c in t.elems.items() if not c <= T.t_id]



def table_flat_for_insert(table):
    pks = {join_names(pk) for pk in table.options.get('pk', [])}
    names = [name for name, t in flatten_type(table)]
    return classify_bool(names, lambda name: name in pks)


def join_names(names):
    return "_".join(names)


# The rest is implemented in display.py
@dp_type
def pql_repr(t, value):
    return repr(value)


def kernel_type(t):
    if t <= T.projected: # or t <= T.aggregated:
        return kernel_type(t.elems['item'])
    return t




from pathlib import Path
from time import time

### XXX Fix for Python 3.8 bug (https://github.com/prompt-toolkit/python-prompt-toolkit/issues/1023)
import asyncio
import selectors
selector = selectors.SelectSelector()
loop = asyncio.SelectorEventLoop(selector)
asyncio.set_event_loop(loop)
### XXX End of fix

from pygments.lexers.go import GoLexer
from pygments.style import Style
from pygments.token import Keyword, Name, Comment, String, Error, Number, Operator, Generic
from prompt_toolkit.styles.pygments import style_from_pygments_cls
from prompt_toolkit import PromptSession
from prompt_toolkit.lexers import PygmentsLexer
from prompt_toolkit.filters import Condition
from prompt_toolkit.validation import Validator, ValidationError
from prompt_toolkit.application.current import get_app
from prompt_toolkit.completion import Completer, Completion
from prompt_toolkit.formatted_text.html import HTML, html_escape
from prompt_toolkit.history import FileHistory
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
from prompt_toolkit.output.color_depth import ColorDepth

from . import __version__
from . import settings
from .utils import memoize
from .loggers import repl_log
from .context import context

from .core.exceptions import Signal, ExitInterp, pql_SyntaxError
from .core.autocomplete import autocomplete
from .core.parser import parse_stmts
from .core import pql_objects as objects
from .core.display import table_more
from .core.pql_types import Object
from .core.pql_types import T


def is_name(s):
    return s.isalnum() or s in ('_', '!')

def last_word(s):
    if not s:
        return '', ''
    i = len(s)
    while i and is_name(s[i-1]):
        i -= 1
    if i < len(s) and s[i] == '!' :
        i += 1  # hack to support ... !var and !in
    return s[:i], s[i:]


class Autocompleter(Completer):
    def __init__(self, state):
        self.state = state

    def _get_completions(self, document):
        previous_text, fragment = last_word(document.text_before_cursor)

        if not settings.autocomplete:
            return

        previous_text = previous_text.rstrip()
        open_complete = previous_text and previous_text[-1] in '.{['
        if not fragment and not open_complete:
            return

        assert open_complete or is_name(fragment[-1]), fragment

        try:
            all_vars = dict(autocomplete(self.state, previous_text))
        except:
            if settings.debug:
                raise
            all_vars = {}

        # all_vars.update(KEYWORDS)
        assert all(isinstance(v, tuple) for v in all_vars.values())
        all_vars = list(all_vars.items())
        all_vars.sort(key=lambda item: (item[1][0], item[0]))

        for k, (_rank, v) in all_vars:
            if k.startswith(fragment):
                a, b = k[:len(fragment)], k[len(fragment):]
                if v is None:
                    t = "<keyword>"
                else:
                    try:
                        t = v.type
                    except AttributeError:
                        t = type(v)

                yield Completion(
                    b, start_position=0,
                    display=HTML('<b>%s</b>%s<ansibrightblack> : %s</ansibrightblack>' % (a, b, html_escape(t))),
                    style='bg:ansigray fg:black',
                    selected_style="fg:black bg:ansibrightyellow",
                    )

    def get_completions(self, document, complete_event):
        return self._get_completions(document)

class MyValidator(Validator):
    def validate(self, document):
        text = document.text
        if not text.strip():
            return
        if text == '.':
            return

        try:
            parse_stmts(text, '<repl>')
        except pql_SyntaxError as e:
            if e.text_ref:
                pos = e.text_ref.ref.end.char_index
                # if pos <= len(text):
                raise ValidationError(message=e.message, cursor_position=pos)
        # except Exception as e:
            # raise ValidationError(message=e.args[0], cursor_position=0)
            # pass

# from prompt_toolkit.key_binding import KeyBindings
# kb = KeyBindings()
# @kb.add('c-space')
# def _(event):
#     " Initialize autocompletion, or select the next completion. "
#     buff = event.app.current_buffer
#     if buff.complete_state:
#         buff.complete_next()
#     else:
#         buff.start_completion(select_first=False)


@memoize
def _code_is_valid(code):
    if code == '.':
        return True
    if code:
        try:
            parse_stmts(code, '<repl>')
        except pql_SyntaxError as e:
            return False
        except Exception as e:
            repl_log.warning(e)

    return True


class PreqlStyle(Style):
    default_style = ""
    styles = {
        Generic:                'ansigray',
        Comment:                'italic #888',
        Keyword:                'bold #00f',
        Name:                   '#fff',
        Name.Function:          'bold #8f8',
        Name.Class:             'bold #0f0',
        String:                 'ansigreen',
        Number:                 'ansicyan',
        Operator:               'ansigray',
        Error:                  'bg:ansired ansigray',
    }


def start_repl(p, prompt=' >> '):
    save_last = '_'   # XXX A little hacky

    p.set_output_format('rich')

    display = p._display
    interp = p._interp
    console = display.console
    console.print(f"[purple]Preql {__version__} interactive prompt. Type help() for help[/purple]")

    try:
        session = PromptSession(
            style=style_from_pygments_cls(PreqlStyle),
            lexer=PygmentsLexer(GoLexer),
            completer=Autocompleter(interp.state),
            # key_bindings=kb
            validator=MyValidator(),
            history=FileHistory(str(Path.home() / '.preql_history')),
            auto_suggest=AutoSuggestFromHistory(),
            color_depth=ColorDepth.TRUE_COLOR,
        )

        @Condition
        def multiline_filter():
            text = get_app().layout.get_buffer_by_name('DEFAULT_BUFFER').text
            return not _code_is_valid(text)

        while True:
            # Read
            try:
                code = session.prompt(prompt, multiline=multiline_filter)
                if not code.strip():
                    continue


                start_time = time()
                try:
                    if code == '.':
                        with context(state=p._interp.state):
                            console.print(table_more(), overflow='ellipsis')
                        continue

                    # Evaluate (Really just compile)
                    res = p._run_code(code, '<repl>')

                    # Print
                    if res is not None and res is not objects.null:
                        assert isinstance(res, Object), (res, type(res))

                        if save_last:
                            p._interp.set_var(save_last, res)

                        with context(state=p._interp.state):
                            res_repr = res.repr()

                        # repl_log.info(res)
                        if isinstance(res_repr, str) and res.type == T.string:  # Not text
                            if len(res_repr) > 200:
                                res_repr = res_repr[:100] + "..." + res_repr[-100:]    # smarter limit?
                        display.print(res_repr)

                except Signal as s:
                    display.print_exception(s)
                    # repl_log.error(s)
                    # p.interp.set_var('_e', objects.ExceptionInstance(e))
                    continue
                except ExitInterp as e:
                    return e.value
                except Exception as e:
                    repl_log.exception(e)
                    raise
                    # continue

                duration = time() - start_time
                if duration > 1:
                    repl_log.info("(Query took %.2f seconds)" % duration)

            except KeyboardInterrupt:
                repl_log.info("Interrupted (Ctrl+C)")



    except (KeyboardInterrupt, EOFError):
        repl_log.info('Exiting Preql interaction')



from typing import List, Optional
import logging
from pathlib import Path

from preql.utils import safezip, dataclass, SafeDict, listgen, method
from preql import settings
from preql.context import context

from .interp_common import assert_type, exclude_fields, call_builtin_func, is_global_scope, cast_to_python_string, cast_to_python_int
from .state import set_var, use_scope, get_var, unique_name, get_db, get_db_target, catch_access, AccessLevels, get_access_level, reduce_access
from . import exceptions as exc
from . import pql_objects as objects
from . import pql_ast as ast
from . import sql
from .parser import Str
from .interp_common import dsp, pyvalue_inst, cast_to_python
from .compiler import cast_to_instance
from .pql_types import T, Type, Object, Id, dp_inst
from .types_impl import table_params, table_flat_for_insert, flatten_type, pql_repr, kernel_type
from .exceptions import InsufficientAccessLevel, ReturnSignal, Signal


MODULES_PATH = Path(__file__).parent.parent / 'modules'


@dsp
def resolve(struct_def: ast.StructDef):
    members = {str(k):resolve(v) for k, v in struct_def.members}
    struct = T.struct(members)
    set_var(struct_def.name, struct)
    return struct

@dsp
def resolve(table_def: ast.TableDef):
    name = table_def.name
    if is_global_scope(context.state):
        name = get_db().qualified_name(name)
        temporary = False
    else:
        name = '__local_' + unique_name(name)
        temporary = True

    t = T.table({}, name=Id(name), temporary=temporary)

    with use_scope({table_def.name: t}):  # For self-reference
        elems = {c.name: resolve(c) for c in table_def.columns}
        t = t(elems)

    if table_def.methods:
        methods = evaluate(table_def.methods)
        t.proto_attrs.update({m.userfunc.name:m.userfunc for m in methods})

    return t

@dsp
def resolve(col_def: ast.ColumnDef):
    coltype = resolve(col_def.type)

    query = col_def.query
    assert not query

    if isinstance(coltype, objects.SelectedColumnInstance):
        table = coltype.parent.type
        if 'name' not in table.options:
            # XXX better test for persistence
            raise Signal.make(T.TypeError, col_def.type, "Tables provided as relations must be persistent.")

        x = T.t_relation[coltype.type](rel={'table': table, 'column': coltype.name, 'key': False})
        return x.replace(_nullable=coltype.type._nullable)  # inherit is_nullable (TODO: use sumtypes?)

    elif coltype <= T.table:
        if 'name' not in coltype.options:
            # XXX better test for persistence
            raise Signal.make(T.TypeError, col_def.type, "Tables provided as relations must be persistent.")

        x = T.t_relation[T.t_id.as_nullable()](rel={'table': coltype, 'column': 'id', 'key': True})
        return x.replace(_nullable=coltype._nullable)     # inherit is_nullable (TODO: use sumtypes?)

    return coltype(default=col_def.default)

@dsp
def resolve(type_: ast.Type):
    t = evaluate( type_.type_obj)
    if isinstance(t, objects.TableInstance):
        t = t.type

    if not isinstance(t, (Type, objects.SelectedColumnInstance)):
        raise Signal.make(T.TypeError, type_, f"Expected type in column definition. Instead got '{t}'")

    if type_.nullable:
        t = t.as_nullable()

    return t



def db_query(sql_code, subqueries=None):
    try:
        return get_db().query(sql_code, subqueries)
    except exc.DatabaseQueryError as e:
        raise Signal.make(T.DbQueryError, None, e.args[0]) from e

def drop_table(state, table_type):
    name ,= table_type.options['name'].parts
    code = sql.compile_drop_table(name)
    return state.db.query(code, {})



@dsp
def _set_value(name: ast.Name, value):
    set_var(name.name, value)

@dsp
def _set_value(attr: ast.Attr, value):
    raise Signal.make(T.NotImplementedError, attr, f"Cannot set attribute for {attr.expr.repr()}")

def _copy_rows(target_name: ast.Name, source: objects.TableInstance):

    if source is objects.EmptyList: # Nothing to add
        return objects.null

    target = evaluate(target_name)

    params = dict(table_params(target.type))
    for p in params:
        if p not in source.type.elems:
            raise Signal.make(T.TypeError, source, f"Missing column '{p}' in {source.type}")

    primary_keys, columns = table_flat_for_insert(target.type)

    source = exclude_fields(source, set(primary_keys) & set(source.type.elems))

    code = sql.Insert(target.type.options['name'], columns, source.code)
    db_query(code, source.subqueries)
    return objects.null


@method
def _execute(struct_def: ast.StructDef):
    resolve(struct_def)

@method
def _execute(table_def: ast.TableDefFromExpr):
    state = context.state
    expr = cast_to_instance(table_def.expr)
    name = table_def.name
    if is_global_scope(state):
        temporary = False
    else:
        name = '__local_' + unique_name(name)
        temporary = True
    t = new_table_from_expr(name, expr, table_def.const, temporary)
    set_var(table_def.name, t)
    
@method
def _execute(var_def: ast.SetValue):
    res = evaluate(var_def.value)
    # res = apply_database_rw(res)
    _set_value(var_def.name, res)
    return res



@method
def _execute(table_def: ast.TableDef):
    if table_def.columns and isinstance(table_def.columns[-1], ast.Ellipsis):
        ellipsis = table_def.columns.pop()
    else:
        ellipsis = None

    if any(isinstance(c, ast.Ellipsis) for c in table_def.columns):
        # XXX why must it? just ensure it appears once
        raise Signal.make(T.SyntaxError, table_def, "Ellipsis must appear at the end")

    # Create type and a corresponding table in the database
    t = resolve(table_def)
    db_name = t.options['name']

    exists = get_db().table_exists(db_name.repr_name)
    if exists:
        assert not t.options['temporary']
        cur_type = get_db().import_table_type(db_name.repr_name, None if ellipsis else set(t.elems) | {'id'})

        if ellipsis:
            elems_to_add = {Str(n, ellipsis.text_ref): v for n, v in cur_type.elems.items() if n not in t.elems}
            # TODO what is primary key isn't included?
            t = t({**t.elems, **elems_to_add}, **cur_type.options)

        # Auto-add id only if it exists already and not defined by user
        if 'id' in cur_type.elems: #and 'id' not in t.elems:
            # t = t(dict(id=T.t_id, **t.elems), pk=[['id']])
            assert cur_type.elems['id'] <= T.primitive, cur_type.elems['id']
            t.elems['id'] = T.t_id

        for e_name, e1_type in t.elems.items():

            if e_name not in cur_type.elems:
                raise Signal.make(T.TypeError, table_def, f"Column '{e_name}' defined, but doesn't exist in database.")

            # e2_type = cur_type.elems[e_name]
            # XXX use can_cast() instead of hardcoding it
            # if not (e1_type <= e2_type or (e1_type <= T.t_id and e2_type <= T.int)):
            #     raise Signal.make(T.TypeError, table_def, f"Cannot cast column '{e_name}' from type '{e2_type}' to '{e1_type}'")

        inst = objects.new_table(t, db_name, select_fields=True)
    else:
        # Auto-add id by default
        elems = dict(t.elems)
        if 'id' not in elems:
            elems = {'id': T.t_id, **elems}
        t = t(elems, pk=[['id']])
        inst = objects.new_table(t, db_name)

    set_var(table_def.name, inst)

    if not exists:
        sql_code = sql.compile_type_def(db_name.repr_name, t)
        db_query(sql_code)
@method
def _execute(insert_rows: ast.InsertRows):
    if not isinstance(insert_rows.name, ast.Name):
        # TODO support Attr
        raise Signal.make(T.SyntaxError, insert_rows, "L-value must be table name")

    rval = evaluate(insert_rows.value)

    assert_type(rval.type, T.table, insert_rows, '+=')

    return _copy_rows(insert_rows.name, rval)

@method
def _execute(func_def: ast.FuncDef):
    func = func_def.userfunc
    assert isinstance(func, objects.UserFunction)

    new_params = []
    for p in func.params:
        if p.type:
            t = evaluate(p.type)
            p = p.replace(type=t)
        new_params.append(p)

    set_var(func.name, func.replace(params=new_params))

@method
def _execute(p: ast.Print):
    display = context.state.display
    # TODO Can be done better. Maybe cast to ReprText?
    insts = evaluate(p.value)
    assert isinstance(insts, list)

    for inst in insts:
        if inst.type <= T.string:
            repr_ = cast_to_python_string(inst)
        else:
            repr_ = inst.repr()

        display.print(repr_, end=" ")
    display.print("")

@method
def _execute(p: ast.Assert):
    res = cast_to_python(p.cond)
    if not res:
        # TODO pretty print values
        if isinstance(p.cond, ast.Compare):
            s = (' %s '%p.cond.op).join(str(evaluate(a).repr()) for a in p.cond.args)
        else:
            s = p.cond.repr()
        raise Signal.make(T.AssertError, p.cond, f"Assertion failed: {s}")

@method
def _execute(cb: ast.CodeBlock):
    for stmt in cb.statements:
        execute(stmt)
    return objects.null


@method
def _execute(i: ast.If):
    cond = cast_to_python(i.cond)

    if cond:
        execute(i.then)
    elif i.else_:
        execute(i.else_)

@method
def _execute(w: ast.While):
    while cast_to_python(w.cond):
        execute(w.do)

@method
def _execute(f: ast.For):
    expr = cast_to_python(f.iterable)
    for i in expr:
        with use_scope({f.var: objects.from_python(i)}):
            execute(f.do)

@method
def _execute(t: ast.Try):
    try:
        execute(t.try_)
    except Signal as e:
        catch_type = evaluate(t.catch_expr).localize()
        if not isinstance(catch_type, Type):
            raise Signal.make(T.TypeError, t.catch_expr, f"Catch expected type, got {t.catch_expr.type}")
        if e.type <= catch_type:
            scope = {t.catch_name: e} if t.catch_name else {}
            with use_scope(scope):
                execute(t.catch_block)
        else:
            raise


def find_module(module_name):
    paths = [MODULES_PATH, Path.cwd()]
    for path in paths:
        module_path =  (path / module_name).with_suffix(".pql")
        if module_path.exists():
            return module_path

    raise Signal.make(T.ImportError, r, "Cannot find module")


def import_module(state, r):
    module_path = find_module(r.module_path)

    # assert state is state.interp.state    # Fix for threaded
    i = state.interp.clone(use_core=r.use_core)

    state.stacktrace.append(r.text_ref)
    try:
        i.include(module_path)
    finally:
        assert state.stacktrace[-1] is r.text_ref
        state.stacktrace.pop()

    # Inherit module db (in case it called connect())
    assert state.db is i.state.db

    ns = i.state.ns
    assert len(ns) == 1
    return objects.Module(r.module_path, ns._ns[0])


@method
def _execute(r: ast.Import):
    module = import_module(context.state, r)
    set_var(r.as_name or r.module_path, module)
    return module

@method
def _execute(r: ast.Return):
    value = evaluate(r.value)
    raise ReturnSignal(value)

@method
def _execute(t: ast.Throw):
    e = evaluate(t.value)
    if isinstance(e, ast.Ast):
        raise exc.InsufficientAccessLevel()
    assert isinstance(e, Exception), e
    raise e

def execute(stmt):
    if isinstance(stmt, ast.Statement):
        return stmt._execute() or objects.null
    return evaluate(stmt)




# Simplify performs local operations before any db-specific compilation occurs
# Technically not super useful at the moment, but makes conceptual sense.

@method
def simplify(cb: ast.CodeBlock):
    # if len(cb.statements) == 1:
    #     s ,= cb.statements
    #     return simplify(s)
    try:
        return cb._execute()
    except ReturnSignal as r:
        # XXX is this correct?
        return r.value
    except Signal as e:
        # Failed to run it, so try to cast as instance
        # XXX order should be other way around!
        if e.type <= T.CastError:
            return cb.compile_to_inst()
        raise
    except InsufficientAccessLevel:
        return cb

@method
def simplify(n: ast.Name):
    # XXX what happens to caching if this is a global variable?
    return get_var(n.name)

@method
def simplify(x: Object):
    return x

# @dsp
# def simplify(ls: list):
#     return [simplify(i) for i in ls]

# @dsp
# def simplify(d: objects.ParamDict):
#     return d.replace(params={name: evaluate( v) for name, v in d.params.items()})

# @dsp
# def simplify(node: ast.Ast):
#     # return _simplify_ast(state, node)
#     return node

# def _simplify_ast(state, node):
#     resolved = {k:simplify(v) for k, v in node
#                 if isinstance(v, types.PqlObject) or isinstance(v, list) and all(isinstance(i, types.PqlObject) for i in v)}
#     return node.replace(**resolved)

# TODO isn't this needed somewhere??
# @dsp
# def simplify(if_: ast.If):
#     if_ = _simplify_ast(state, if_)
#     if isinstance(if_.cond, objects.ValueInstance): # XXX a more general test?
#         if if_.cond.local_value:
#             return if_.then
#         else:
#             return if_.else_
#     return if_


# TODO Optimize these, right now failure to evaluate will lose all work
@method
def simplify(obj: ast.Or):
    a, b = evaluate( obj.args)
    ta = kernel_type(a.type)
    tb = kernel_type(b.type)
    if ta != tb:
        raise Signal.make(T.TypeError, obj, f"'or' operator requires both arguments to be of the same type, but got '{ta}' and '{tb}'.")
    try:
        if a.test_nonzero():
            return a
    except InsufficientAccessLevel:
        return obj
    return b


@method
def simplify(obj: ast.And):
    a, b = evaluate( obj.args)
    ta = kernel_type(a.type)
    tb = kernel_type(b.type)
    if ta != tb:
        raise Signal.make(T.TypeError, obj, f"'and' operator requires both arguments to be of the same type, but got '{ta}' and '{tb}'.")
    try:
        if not a.test_nonzero():
            return a
    except InsufficientAccessLevel:
        return obj
    return b


@method
def simplify(obj: ast.Not):
    inst = evaluate( obj.expr)
    try:
        nz = inst.test_nonzero()
    except InsufficientAccessLevel:
        return obj
    return objects.pyvalue_inst(not nz)



@method
def simplify(funccall: ast.FuncCall):
    state = context.state
    func = evaluate(funccall.func)

    if isinstance(func, objects.UnknownInstance):
        # evaluate( [a.value for a in funccall.args])
        raise Signal.make(T.TypeError, funccall.func, f"Error: Object of type '{func.type}' is not callable")

    args = funccall.args
    if isinstance(func, Type):
        # Cast to type
        args = args + [func]
        func = get_var('cast')

    if not isinstance(func, objects.Function):
        raise Signal.make(T.TypeError, funccall.func, f"Error: Object of type '{func.type}' is not callable")

    state.stacktrace.append(funccall.text_ref)
    try:
        res = eval_func_call(func, args)
    finally:
        assert state.stacktrace[-1] is funccall.text_ref
        state.stacktrace.pop()

    assert isinstance(res, Object), (type(res), res)
    return res


def eval_func_call(func, args):
    state = context.state
    assert isinstance(func, objects.Function)

    matched_args = func.match_params(args)

    if isinstance(func, objects.MethodInstance):
        ordered_args = {'this': func.parent}
        func = func.func
        # args.update(func.parent.all_attrs())
    else:
        ordered_args = {}

    # XXX simplify destroys text_ref, so it harms error messages.
    # TODO Can I get rid of it, or make it preserve the text_ref somehow?
    # Don't I need an instance to ensure I have type?

    for i, (p, a) in enumerate(matched_args):
        if not p.name.startswith('$'):      # $param means don't evaluate expression, leave it to the function
            a = evaluate( a)
        # TODO cast?
        if p.type and not a.type <= p.type:
            raise Signal.make(T.TypeError, func, f"Argument #{i} of '{func.name}' is of type '{a.type}', expected '{p.type}'")
        ordered_args[p.name] = a


    if isinstance(func, objects.InternalFunction):
        # TODO ensure pure function?
        # TODO Ensure correct types
        ordered_args = list(ordered_args.values())
        return func.func(*ordered_args)

    # TODO make tests to ensure caching was successful
    expr = func.expr
    if settings.cache:
        params = {name: ast.Parameter(name, value.type) for name, value in ordered_args.items()}
        sig = (func.name,) + tuple(a.type for a in ordered_args.values())

        try:
            with use_scope(params):
                if sig in state._cache:
                    compiled_expr = state._cache[sig]
                else:
                    logging.info(f"Compiling.. {func}")
                    with context(state=reduce_access(AccessLevels.COMPILE)):
                        compiled_expr = _call_expr(func.expr)
                    logging.info("Compiled successfully")
                    if isinstance(compiled_expr, objects.Instance):
                        # XXX a little ugly
                        qb = sql.QueryBuilder(get_db_target(), True)
                        x = compiled_expr.code.compile(qb)
                        x = x.optimize()
                        compiled_expr = compiled_expr.replace(code=x)
                    state._cache[sig] = compiled_expr

            expr = ast.ResolveParameters(compiled_expr, ordered_args)

        except exc.InsufficientAccessLevel:
            # Don't cache
            pass

    with use_scope(ordered_args):
        res = _call_expr(expr)

    if isinstance(res, ast.ResolveParameters):  # XXX A bit of a hack
        raise exc.InsufficientAccessLevel()

    return res


def _call_expr(expr):
    try:
        return evaluate(expr)
    except ReturnSignal as r:
        return r.value

# TODO fix these once we have proper types
@method
def test_nonzero(table: objects.TableInstance):
    count = call_builtin_func("count", [table])
    return bool(cast_to_python_int(count))

@method
def test_nonzero(inst: objects.Instance):
    return bool(cast_to_python(inst))

@method
def test_nonzero(inst: Type):
    return True








@method
def apply_database_rw(o: ast.One):
    # TODO move these to the core/base module
    obj = evaluate( o.expr)
    if obj.type <= T.struct:
        if len(obj.attrs) != 1:
            raise Signal.make(T.ValueError, o, f"'one' expected a struct with a single attribute, got {len(obj.attrs)}")
        x ,= obj.attrs.values()
        return x

    slice_ast = ast.Slice(obj, ast.Range(None, ast.Const(T.int, 2))).set_text_ref(o.text_ref)
    table = evaluate( slice_ast)

    assert (table.type <= T.table), table
    rows = table.localize() # Must be 1 row
    if len(rows) == 0:
        if not o.nullable:
            raise Signal.make(T.ValueError, o, "'one' expected a single result, got an empty expression")
        return objects.null
    elif len(rows) > 1:
        raise Signal.make(T.ValueError, o, "'one' expected a single result, got more")

    row ,= rows
    rowtype = T.row[table.type]

    if table.type <= T.list:
        return pyvalue_inst(row)

    assert table.type <= T.table
    assert_type(table.type, T.table, o, 'one')
    d = {k: pyvalue_inst(v, table.type.elems[k], True) for k, v in row.items()}
    return objects.RowInstance(rowtype, d)


@method
def apply_database_rw(d: ast.Delete):
    catch_access(AccessLevels.WRITE_DB)
    # TODO Optimize: Delete on condition, not id, when possible

    cond_table = ast.Selection(d.table, d.conds).set_text_ref(d.text_ref)
    table = evaluate( cond_table)

    if not table.type <= T.table:
        raise Signal.make(T.TypeError, d.table, f"Expected a table. Got: {table.type}")

    if not 'name' in table.type.options:
        raise Signal.make(T.ValueError, d.table, "Cannot delete. Table is not persistent")

    rows = list(table.localize())
    if rows:
        if 'id' not in rows[0]:
            raise Signal.make(T.TypeError, d, "Delete error: Table does not contain id")

        ids = [row['id'] for row in rows]

        for code in sql.deletes_by_ids(table, ids):
            db_query(code, table.subqueries)

    return evaluate( d.table)

@method
def apply_database_rw(u: ast.Update):
    catch_access(AccessLevels.WRITE_DB)

    # TODO Optimize: Update on condition, not id, when possible
    table = evaluate( u.table)

    if not table.type <= T.table:
        raise Signal.make(T.TypeError, u.table, f"Expected a table. Got: {table.type}")

    if not 'name' in table.type.options:
        raise Signal.make(T.ValueError, u.table, "Cannot update: Table is not persistent")

    for f in u.fields:
        if not f.name:
            raise Signal.make(T.SyntaxError, f, f"Update requires that all fields have a name")

    # TODO verify table is concrete (i.e. lvalue, not a transitory expression)

    update_scope = {n:c for n, c in table.all_attrs().items()}
    with use_scope(update_scope):
        proj = {f.name:evaluate( f.value) for f in u.fields}

    rows = list(table.localize())
    if rows:
        if 'id' not in rows[0]:
            raise Signal.make(T.TypeError, u, "Update error: Table does not contain id")
        if not set(proj) < set(rows[0]):
            raise Signal.make(T.TypeError, u, "Update error: Not all keys exist in table")

        ids = [row['id'] for row in rows]

        for code in sql.updates_by_ids(table, proj, ids):
            db_query(code, table.subqueries)

    # TODO return by ids to maintain consistency, and skip a possibly long query
    return table


@method
def apply_database_rw(new: ast.NewRows):
    catch_access(AccessLevels.WRITE_DB)

    obj = get_var(new.type)

    if len(new.args) > 1:
        raise Signal.make(T.NotImplementedError, new, "Not yet implemented") #. Requires column-wise table concat (use join and enum)")

    if isinstance(obj, objects.UnknownInstance):
        arg ,= new.args
        table = evaluate( arg.value)
        fakerows = [objects.RowInstance(T.row[table], {'id': T.t_id})]
        return ast.List_(T.list[T.int], fakerows).set_text_ref(new.text_ref)

    if isinstance(obj, objects.TableInstance):
        # XXX Is it always TableInstance? Just sometimes? What's the transition here?
        obj = obj.type

    assert_type(obj, T.table, new, "'new' expected an object of type '%s', instead got '%s'")

    arg ,= new.args

    # TODO postgres can do it better!
    table = evaluate( arg.value)
    rows = table.localize()

    # TODO ensure rows are the right type

    cons = TableConstructor.make(obj)

    # TODO very inefficient, vectorize this
    ids = []
    for row in rows:
        matched = cons.match_params([objects.from_python(v) for v in row.values()])
        ids += [_new_row(new, obj, matched).primary_key()]   # XXX return everything, not just pk?

    # XXX find a nicer way - requires a better typesystem, where id(t) < int
    return ast.List_(T.list[T.int], ids).set_text_ref(new.text_ref)


@listgen
def _destructure_param_match(ast_node, param_match):
    # TODO use cast rather than a ad-hoc hardwired destructure
    for k, v in param_match:
        if isinstance(v, objects.RowInstance):
            v = v.primary_key()
        v = v.localize()

        if k.type <= T.struct:
            names = [name for name, t in flatten_type(k.orig, [k.name])]
            if not isinstance(v, list):
                msg = f"Parameter {k.name} received a bad value: {v} (expecting a struct or a list)"
                raise Signal.make(T.TypeError, ast_node, msg)
            if len(v) != len(names):
                msg = f"Parameter {k.name} received a bad value (size of {len(names)})"
                raise Signal.make(T.TypeError, ast_node, msg)
            yield from safezip(names, v)
        else:
            yield k.name, v


def _new_value(v, type_):
    if isinstance(v, list):
        return evaluate( objects.PythonList(v))
    return objects.pyvalue_inst(v, type_=type_)

@dsp
def freeze(i: objects.Instance):
    return _new_value(cast_to_python(i), type_=i.type )

@dsp
def freeze(i: objects.RowInstance):
    return i.replace(attrs={k: freeze(v) for k, v in i.attrs.items()})

def _new_row(new_ast, table, matched):
    matched = [(k, freeze(evaluate( v))) for k, v in matched]
    destructured_pairs = _destructure_param_match(new_ast, matched)

    keys = [name for (name, _) in destructured_pairs]
    values = [sql.make_value(v) for (_,v) in destructured_pairs]
    assert keys and values
    # XXX use regular insert?

    if 'name' not in table.options:
        raise Signal.make(T.TypeError, new_ast, f"'new' expects a persistent table. Instead got a table expression.")

    if get_db_target() == sql.bigquery:
        rowid = db_query(sql.FuncCall(T.string, 'GENERATE_UUID', []))
        keys += ['id']
        values += [sql.make_value(rowid)]
        q = sql.InsertConsts(table.options['name'].repr_name, keys, [values])
        db_query(q)
    else:
        q = sql.InsertConsts(table.options['name'].repr_name, keys, [values])
        # q = sql.InsertConsts(new_ast.type, keys, [values])
        db_query(q)
        rowid = db_query(sql.LastRowId())

    d = SafeDict({'id': objects.pyvalue_inst(rowid)})
    d.update({p.name:v for p, v in matched})
    return objects.RowInstance(T.row[table], d)



@method
def apply_database_rw(new: ast.New):
    catch_access(AccessLevels.WRITE_DB)

    obj = get_var(new.type)

    # XXX Assimilate this special case
    if isinstance(obj, Type) and obj <= T.Exception:
        def create_exception(msg):
            state = context.state
            msg = cast_to_python(msg)
            assert new.text_ref is state.stacktrace[-1]
            return Signal(obj, list(state.stacktrace), msg)    # TODO move this to `throw`?
        f = objects.InternalFunction(obj.typename, [objects.Param('message')], create_exception)
        res = evaluate( ast.FuncCall(f, new.args).set_text_ref(new.text_ref))
        return res

    if not isinstance(obj, objects.TableInstance):
        raise Signal.make(T.TypeError, new, f"'new' expects a table or exception, instead got {obj.repr()}")

    table = obj
    # TODO assert tabletype is a real table and not a query (not transient), otherwise new is meaningless
    assert_type(table.type, T.table, new, "'new' expected an object of type '%s', instead got '%s'")

    cons = TableConstructor.make(table.type)
    matched = cons.match_params(new.args)

    return _new_row(new, table.type, matched)

@method
def apply_database_rw(x: Object):
    return x


@dataclass
class TableConstructor(objects.Function):
    "Serves as an ad-hoc constructor function for given table, to allow matching params"

    params: List[objects.Param]
    param_collector: Optional[objects.Param] = None
    name = 'new'

    @classmethod
    def make(cls, table):
        return cls([
            objects.Param(name, p, p.options.get('default'), orig=p).set_text_ref(getattr(name, 'text_ref', None))
            for name, p in table_params(table)
        ])


def add_as_subquery(inst: objects.Instance):
    code_cls = sql.TableName if (inst.type <= T.table) else sql.Name
    name = unique_name(inst)
    return inst.replace(code=code_cls(inst.code.type, name), subqueries=inst.subqueries.update({name: inst.code}))


@method
def resolve_parameters(x: Object):
    return x

@method
def resolve_parameters(p: ast.Parameter):
    return get_var(p.name)


@dsp
def evaluate(obj: list):
    return [evaluate( item) for item in obj]

@dsp
def evaluate(obj_):
    access_level = get_access_level()

    # - Generic, non-db related operations
    obj = obj_.simplify()
    assert obj, obj_

    if access_level < AccessLevels.COMPILE:
        return obj

    # - Compile to instances with db-specific code (sql)
    # . Compilation may fail (e.g. due to lack of DB access)
    # . Resulting code generic within the same database, and can be cached
    # obj = compile_to_inst(state.reduce_access(state.AccessLevels.COMPILE), obj)
    obj = obj.compile_to_inst()

    if access_level < AccessLevels.EVALUATE:
        return obj

    # - Resolve parameters to "instantiate" the cached code
    # TODO necessary?
    obj = obj.resolve_parameters()

    if access_level < AccessLevels.READ_DB:
        return obj

    # - Apply operations that read or write the database (delete, insert, update, one, etc.)
    obj = obj.apply_database_rw()

    assert not isinstance(obj, (ast.ResolveParameters, ast.ParameterizedSqlCode)), obj

    return obj



#
#    localize()
# -------------
#
# Return the local value of the expression. Only requires computation if the value is an instance.
#

@method
def localize(inst: objects.AbsInstance):
    raise NotImplementedError(inst)

@method
def localize(inst: objects.AbsStructInstance):
    return {k: evaluate(v).localize() for k, v in inst.attrs.items()}

@method
def localize(inst: objects.Instance):
    # TODO This protection doesn't work for unoptimized code
    # Cancel unoptimized mode? Or leave this unprotected?
    # state.require_access(state.AccessLevels.WRITE_DB)

    if inst.code is sql.null:
        return None

    return db_query(inst.code, inst.subqueries)

@method
def localize(inst: objects.ValueInstance):
    return inst.local_value

@method
def localize(inst: objects.SelectedColumnInstance):
    # XXX is this right?
    p = evaluate(inst.parent)
    return p.get_attr(inst.name)

@method
def localize(x: Object):
    return x





def new_table_from_rows(name, columns, rows):
    # TODO check table doesn't exist

    tuples = [
        [sql.make_value(i) for i in row]
        for row in rows
    ]

    # TODO refactor into function?
    elems = {c:v.type.as_nullable() for c,v in zip(columns, tuples[0])}
    elems['id'] = T.t_id
    table = T.table(elems, temporary=True, pk=[['id']], name=Id(name))

    db_query(sql.compile_type_def(name, table))

    code = sql.InsertConsts(name, columns, tuples)
    db_query(code)

    x = objects.new_table(table)
    set_var(name, x)
    return x


def new_table_from_expr(name, expr, const, temporary):
    elems = expr.type.elems

    if any(t <= T.unknown for t in elems.values()):
        return objects.TableInstance.make(sql.null, expr.type, [])

    if 'id' in elems and not const:
        msg = "Field 'id' already exists. Rename it, or use 'const table' to copy it as-is."
        raise Signal.make(T.NameError, None, msg)

    table = T.table(dict(elems), name=Id(name), pk=[] if const else [['id']], temporary=temporary)

    if not const:
        table.elems['id'] = T.t_id

    db_query(sql.compile_type_def(name, table))

    read_only, flat_columns = table_flat_for_insert(table)
    expr = exclude_fields(expr, set(read_only) & set(elems))
    db_query(sql.Insert(Id(name), flat_columns, expr.code), expr.subqueries)

    return objects.new_table(table)


# cast_to_python - make sure the value is a native python object, not a preql instance

@dsp
def cast_to_python(obj):
    raise Signal.make(T.TypeError, None, f"Unexpected value: {pql_repr(obj.type, obj)}")

@dsp
def cast_to_python(obj: ast.Ast):
    inst = cast_to_instance(obj)
    return cast_to_python(inst)

@dsp
def cast_to_python(obj: objects.AbsInstance):
    # if state.access_level <= state.AccessLevels.QUERY:
    if obj.type <= T.projected | T.aggregated:
        raise exc.InsufficientAccessLevel(get_access_level())
        # raise Signal.make(T.CastError, None, f"Internal error. Cannot cast projected obj: {obj}")
    res = obj.localize()
    if obj.type == T.float:
        res = float(res)
    elif obj.type == T.int:
        res = int(res)
    elif obj.type == T.bool:
        assert res in (0, 1), res
        res = bool(res)
    return res



### Added functions

def function_localize_keys(self, struct):
    return cast_to_python(struct)

objects.Function._localize_keys = function_localize_keys


def instance_repr(self):
    return pql_repr(self.type, self.localize())

objects.Instance.repr = instance_repr


@dp_inst
def post_instance_getattr(inst, p: T.property):
    return eval_func_call(objects.MethodInstance(inst, p.func), [])



import html

import rich.table
import rich.text
import rich.console

from .exceptions import Signal
from .pql_types import T, ITEM_NAME
from . import pql_objects as objects
from .pql_ast import pyvalue
from .types_impl import dp_type, pql_repr
from .interp_common import call_builtin_func, cast_to_python_int, cast_to_python
from .state import get_display


TABLE_PREVIEW_SIZE = 16
LIST_PREVIEW_SIZE = 128
MAX_AUTO_COUNT = 10000

@dp_type
def pql_repr(t: T.function, value):
    params = []
    for p in value.params:
        s = p.name
        if p.type:
            s += ": %s" % p.type
        params.append(s)

    return f'{{func {value.name}({", ".join(params)})}}'

@dp_type
def pql_repr(t: T.decimal, value):
    raise Signal.make(T.NotImplementedError, None, "Decimal not implemented")

@dp_type
def pql_repr(t: T.string, value):
    if not isinstance(value, str):
        raise Signal.make(T.TypeError, None, f"Expected value of type 'string', instead got {type(value)}")

    value = value.replace('"', r'\"')
    res = f'"{value}"'
    if get_display().format == 'html':
        res = html.escape(res)
    return res

@dp_type
def pql_repr(t: T.text, value):
    assert isinstance(value, str), value
    return str(value)

@dp_type
def pql_repr(t: T._rich, value):
    r = rich.text.Text.from_markup(str(value))
    if get_display().format == 'html':
        return _rich_to_html(r)
    return r

@dp_type
def pql_repr(t: T.bool, value):
    return 'true' if value else 'false'

@dp_type
def pql_repr(t: T.nulltype, value):
    return 'null'


def _rich_to_html(r):
    console = rich.console.Console(record=True)
    console.print(r)
    return console.export_html(code_format='<style>{stylesheet}</style><pre>{code}</pre>').replace('━', '-')


def table_limit(table, limit, offset=0):
    assert isinstance(limit, int)
    assert isinstance(offset, int)
    return call_builtin_func('limit_offset', [table, pyvalue(limit), pyvalue(offset)])


def _html_table(name, count_str, rows, offset, has_more, colors):
    assert colors
    header = 'table '
    if name:
        header += name
    if offset:
        header += f'[{offset}..]'
    header += f" {count_str}"
    header = f"<pre>table {name}, {count_str}</pre>"

    if not rows:
        return header

    cols = list(rows[0])
    ths = '<tr>%s</tr>' % ' '.join([f"<th>{col}</th>" for col in cols])
    trs = [
        '<tr>%s</tr>' % ' '.join([f"<td>{v}</td>" for v in row.values()])
        for row in rows
    ]

    if has_more:
        trs.append('<tr><td>...</td></tr>')

    style = """<style>
    .preql_table td, .preql_table th {
        text-align: left
    }
    </style>
    """
    return '%s<table class="preql_table">%s%s</table>' % (header, ths, '\n'.join(trs)) + style


def _rich_table(name, count_str, rows, offset, has_more, colors=True, show_footer=False):
    header = 'table '
    if name:
        header += name
    if offset:
        header += f'[{offset}..]'
    header += f" {count_str}"

    if not rows:
        return header

    table = rich.table.Table(title=rich.text.Text(header), show_footer=show_footer, min_width=len(header))

    # TODO enable/disable styling
    for k, v in rows[0].items():
        kw = {}
        if isinstance(v, (int, float)):
            kw['justify']='right'

        if colors:
            if isinstance(v, int):
                kw['style']='cyan'
            elif isinstance(v, float):
                kw['style']='yellow'
            elif isinstance(v, str):
                kw['style']='green'

        table.add_column(k, footer=k, **kw)

    for r in rows:
        table.add_row(*[rich.markup.escape(str(x) if x is not None else '-') for x in r.values()])

    if has_more:
        table.add_row(*['...' for x in rows[0]])

    return table

_g_last_table = None
_g_last_offset = 0

def _view_table(table, size, offset):
    global _g_last_table, _g_last_offset
    rows = cast_to_python(table_limit(table, size, offset))
    _g_last_table = table
    _g_last_offset = offset + len(rows)
    if table.type <= T.list:
        rows = [{ITEM_NAME: x} for x in rows]

    try:
        table_name = table.type.options['name'].repr_name
    except KeyError:
        table_name = ''

    return table_name, rows


def table_inline_repr(self):
    offset = 0
    table_name, rows, = _view_table(self, TABLE_PREVIEW_SIZE, offset)
    return '[%s]' % ', '.join(repr(r) for r in rows)



def table_repr(self, offset=0):

    count = cast_to_python_int(call_builtin_func('count', [table_limit(self, MAX_AUTO_COUNT)]))
    if count == MAX_AUTO_COUNT:
        count_str = f'>={count}'
    else:
        count_str = f'={count}'

    # if len(self.type.elems) == 1:
    #     rows = cast_to_python(table_limit(self, LIST_PREVIEW_SIZE))
    #     post = f', ... ({count_str})' if len(rows) < count else ''
    #     elems = ', '.join(repr_value(ast.Const(None, self.type.elem, r)) for r in rows)
    #     return f'[{elems}{post}]'

    # TODO load into preql and repr, instead of casting to python
    table_f = _rich_table
    preview = TABLE_PREVIEW_SIZE
    colors = True
    display = get_display()

    if display.format == 'html':
        preview = TABLE_PREVIEW_SIZE * 10
        table_f = _html_table
    elif display.format == 'text':
        colors = False
    else:
        assert display.format == 'rich'

    table_name, rows, = _view_table(self, preview, offset)
    has_more = offset + len(rows) < count
    return table_f(table_name, count_str, rows, offset, has_more, colors=colors)


def table_more():
    if not _g_last_table:
        raise Signal.make(T.ValueError, None, "No table yet")

    return table_repr(_g_last_table, _g_last_offset)


def module_repr(module):
    res = f'<Module {module.name} | {len(module.namespace)} members>'
    if get_display().format == 'html':
        res = html.escape(res)
    return res

def function_repr(func):
    res = '<%s>' % func.help_str()
    if get_display().format == 'html':
        res = html.escape(res)
    return res


class Display:
    def print(self, repr_):
        print(repr_)

def _print_rich_exception(console, e):
    console.print('[bold]Exception traceback:[/bold]')
    for ref in e.text_refs:
        for line in (ref.get_pinpoint_text(rich=True) if ref else ['???']):
            console.print(line)
        console.print()
    console.print(rich.text.Text('%s: %s' % (e.type, e.message)))

class RichDisplay(Display):
    format = "rich"

    def __init__(self):
        self.console = rich.console.Console()

    def print(self, repr_, end="\n"):
        if hasattr(repr_, '__rich_console__'):
            self.console.print(repr_, overflow="ellipsis", end=end)
        else:
            # repr_ = rich.text.Text(repr_)
            self.console.print(repr_, overflow="ellipsis", markup=False, end=end)

    def print_exception(self, e):
        "Yields colorful styled lines to print by the ``rich`` library"
        _print_rich_exception(self.console, e)


def print_to_string(x, format):
    console = rich.console.Console(color_system=None)
    with console.capture() as capture: 
        console.print(x)
    return capture.get()


class HtmlDisplay(Display):
    format = "html"

    def __init__(self):
        # self.console = rich.console.Console(record=True)
        self.buffer = []

    def print(self, repr_, end="<br/>"):
        self.buffer.append(str(repr_) + end)

    # print = RichDisplay.print
    def print_exception(self, e):
        console = rich.console.Console(record=True)
        _print_rich_exception(console, e)
        res = console.export_html(code_format='<style>{stylesheet}</style><pre>{code}</pre>').replace('━', '-')
        self.buffer.append(res)

    def as_html(self):
        # return '\n'.join(self.buffer) + "%%"
        # return self.console.export_html(code_format='<style>{stylesheet}</style><pre>{code}</pre>').replace('━', '-')
        res = '\n'.join(self.buffer)
        self.buffer.clear()
        return res



def install_reprs():
    objects.CollectionInstance.repr = table_repr
    objects.CollectionInstance.inline_repr = table_inline_repr
    objects.Module.repr = module_repr
    objects.Function.repr = function_repr




class Object:
    def repr(self):
        return repr(self)

    def inline_repr(self):
    	return self.repr()

    def rich_repr(self):
        return self.repr().replace('[', '\\[')


    def all_attrs(self):
        return {}

    get_attr = NotImplemented
    isa = NotImplemented



import threading
from contextlib import contextmanager

class Context(threading.local):
    def __init__(self):
        self._ctx = [{}]

    def __getattr__(self, name):
        for scope in reversed(self._ctx):
            if name in scope:
                return scope[name]

        raise AttributeError(name)

    def get(self, name, default=None):
        try:
            return getattr(self, name)
        except AttributeError:
            return default

    @contextmanager
    def __call__(self, **attrs):
        self._ctx.append(attrs)
        try:
            yield
        finally:
            _d = self._ctx.pop()
            assert attrs is _d



def test_threading():
    import time, random

    context = Context()

    def f(i):
        with context(i=i):
            g(i)

    def g(i):
        assert context.i == i
        time.sleep(random.random())
        assert context.i == i
        print(i, end=', ')


    for i in range(100):
        t = threading.Thread(target=f, args=(i,))
        t.start()


# test_threading()

context = Context()



"""This module provides functions for importing the return values from SQL queries.

The value are already "native python", but not conforming to the expected type.

We perform two operations:
- Place the items into the expected structure (instead of just a flat list)
- Convert primitives to the expected type

"""

from datetime import datetime
import decimal
import json

import arrow

from preql.utils import listgen, safezip
from .pql_types import T, dp_type, dp_inst
from .exceptions import Signal
from .types_impl import flatten_type
from preql.context import context
from .sql import _ARRAY_SEP, sqlite, mysql


def _from_datetime(s):
    if s is None:
        return None

    # Postgres
    if isinstance(s, datetime):
        return s

    # Sqlite
    if isinstance(s, str):
        try:
            # return datetime.fromisoformat(s)
            return arrow.get(s)
        except ValueError as e:
            raise Signal.make(T.ValueError, None, str(e))

    raise Signal.make(T.TypeError, None, f"Unexpected type for datetime: {type(s)}")


@dp_type
def _restructure_result(t, i):
    raise Signal.make(T.TypeError, None, f"Unexpected type used: {t}")

@dp_type
def _restructure_result(t: T.table, i):
    # return ({name: _restructure_result(state, col, i) for name, col in t.elem_dict.items()})
    return next(i)

@dp_type
def _restructure_result(t: T.struct, i):
    return ({name: _restructure_result(col, i) for name, col in t.elems.items()})

@dp_type
def _restructure_result(_t: T.union[T.primitive, T.nulltype], i):
    return next(i)


@dp_type
def _restructure_result(t: T.json_array[T.union[T.primitive, T.nulltype]], i):
    res = next(i)
    if not res:
        return res

    target = context.state.db.target
    if target == mysql:
        res = json.loads(res)
    elif target == sqlite:
        if not isinstance(res, str):
            raise Signal.make(T.TypeError, None, f"json_array type expected a string separated by {_ARRAY_SEP}. Got: '{res}'")
        res = res.split(_ARRAY_SEP)

    # XXX hack! TODO Use a generic form to cast types
    try:
        if t.elem <= T.int:
            res = [int(x) for x in res]
        elif t.elem <= T.float:
            res = [float(x) for x in res]
    except ValueError:
        raise Signal.make(T.TypeError, None, f"Error trying to convert values to type {t.elem}")

    return res

@dp_type
def _restructure_result(_t: T.datetime, i):
    s = next(i)
    return _from_datetime(s)



@dp_inst
def sql_result_to_python(res: T.primitive):
    try:
        row ,= res.value
        item ,= row
    except ValueError:
        raise Signal.make(T.TypeError, None, "Expected primitive. Got: '%s'" % res.value)
    # t = from_python(type(item))
    # if not (t <= res.type):
    #     raise Signal.make(T.TypeError, state, None, f"Incorrect type returned from SQL: '{t}' instead of '{res.type}'")
    return item


@dp_inst
def sql_result_to_python(res: T.datetime):
    # XXX doesn't belong here?
    row ,= res.value
    item ,= row
    s = item
    return _from_datetime(s)

def _from_sql_primitive(p):
    if isinstance(p, decimal.Decimal):
        # TODO Needs different handling when we expect a decimal
        return float(p)
    return p

@dp_inst
def sql_result_to_python(arr: T.list):
    fields = flatten_type(arr.type)
    if not all(len(e)==len(fields) for e in arr.value):
        raise Signal.make(T.TypeError, None, f"Expected 1 column. Got {len(arr.value[0])}")

    if arr.type.elem <= T.struct:
        return [{n: _from_sql_primitive(e) for (n, _t), e in safezip(fields, tpl)} for tpl in arr.value]
    else:
        return [_from_sql_primitive(e[0]) for e in arr.value]

@dp_inst
@listgen
def sql_result_to_python(arr: T.table):
    expected_length = len(flatten_type(arr.type))   # TODO optimize?
    for row in arr.value:
        if len(row) != expected_length:
            raise Signal.make(T.TypeError, None, f"Expected {expected_length} columns, but got {len(row)}")
        i = iter(row)
        yield {name: _restructure_result(col, i) for name, col in arr.type.elems.items()}





def _bool_from_sql(n):
    if n == 'NO':
        n = False
    if n == 'YES':
        n = True
    assert isinstance(n, bool), n
    return n

def type_from_sql(type, nullable):
    type = type.lower()
    d = {
        'integer': T.int,
        'int': T.int,           # mysql
        'tinyint(1)': T.bool,   # mysql
        'serial': T.t_id,
        'bigserial': T.t_id,
        'smallint': T.int,  # TODO smallint / bigint?
        'bigint': T.int,
        'character varying': T.string,
        'character': T.string,  # TODO char?
        'real': T.float,
        'float': T.float,
        'double precision': T.float,    # double on 32-bit?
        'boolean': T.bool,
        'timestamp': T.datetime,
        'timestamp without time zone': T.datetime,
        'text': T.text,
    }
    try:
        v = d[type]
    except KeyError:
        if type.startswith('int('): # TODO actually parse it
            return T.int
        elif type.startswith('tinyint('): # TODO actually parse it
            return T.int
        elif type.startswith('varchar('): # TODO actually parse it
            return T.string

        return T.string.as_nullable()

    nullable = _bool_from_sql(nullable)

    return v.replace(_nullable=nullable)



import argparse
from pathlib import Path
from itertools import chain
import time

from . import Preql, __version__, Signal

parser = argparse.ArgumentParser(description='Preql command-line interface (aka REPL)')
parser.add_argument('-i', '--interactive', action='store_true', default=False,
                    help="enter interactive mode after running the script")
parser.add_argument('-v', '--version', action='store_true', help="print version")
parser.add_argument('--install-jupyter', action='store_true', help="installs the Preql plugin for Jupyter notebook")
parser.add_argument('--print-sql', action='store_true', help="print the SQL code that's being executed")
parser.add_argument('-f', '--file', type=str, help='path to a Preql script to run')
parser.add_argument('-m', '--module', type=str, help='name of a Preql module to run')
parser.add_argument('--time', action='store_true', help='displays how long the script ran')
parser.add_argument('database', type=str, nargs='?', default=None,
                    help="database url (postgres://user:password@host:port/db_name")
parser.add_argument('--python-traceback', action='store_true',
                    help="Show the Python traceback when an exception causes the interpreter to quit")


def find_dot_preql():
    cwd = Path.cwd()
    for p in chain([cwd], cwd.parents):
        dot_preql = p / ".preql"
        if dot_preql.exists():
            return dot_preql

def main():
    args = parser.parse_args()

    if args.version:
        print(__version__)

    if args.install_jupyter:
        from .jup_kernel.install import main as install_jupyter
        install_jupyter([])
        print("Install successful. To start working, run 'jupyter notebook' and create a new Preql notebook.")

    kw = {'print_sql': args.print_sql}
    if args.database:
        kw['db_uri'] = args.database
        kw['auto_create'] = True
    p = Preql(**kw)

    interactive = args.interactive

    res = 0
    start = time.time()
    try:
        if args.file:
            p.load(args.file)
        elif args.module:
            p('import ' + args.module)
        elif args.version or args.install_jupyter:
            pass
        else:
            dot_preql = find_dot_preql()
            if dot_preql:
                print("Auto-running", dot_preql)
                p._run_code(dot_preql.read_text(), dot_preql)

            interactive = True
    except Signal as e:
        p._display.print_exception(e)
        res = -1
        if args.python_traceback:
            raise
    except KeyboardInterrupt:
        print("Interrupted (Ctrl+C)")

    end = time.time()
    if args.time:
        print('Script took %.2f seconds to run' % (end -start))

    if interactive:
        p.load_all_tables()
        p.start_repl()
    else:
        return res

if __name__ == '__main__':
    main()


from logging import getLogger
from copy import copy
from contextlib import contextmanager

from preql.context import context


from .exceptions import InsufficientAccessLevel, Signal
from .pql_types import T
from . import pql_ast as ast

class NameNotFound(Exception):
    pass


logger = getLogger('state')

class Namespace:
    def __init__(self, ns=None):
        self._ns = ns or [{}]
        # self._parameters = None

    def __copy__(self):
        return Namespace([self._ns[0]] + [dict(n) for n in self._ns[1:]])     # Shared global namespace

    def get_var(self, name):
        # Uses context.parameters, set in Interpreter
        parameters = getattr(context, 'parameters', None)
        if parameters and name in parameters:
            return parameters[name]

        for scope in reversed(self._ns):
            if name in scope:
                return scope[name]

        raise NameNotFound(name)

    def set_var(self, name, value):
        assert not isinstance(value, ast.Name)
        self._ns[-1][name] = value


    @contextmanager
    def use_scope(self, scope: dict):
        x = len(self._ns)
        self._ns.append(scope)
        try:
            yield
        finally:
            _discarded_scope = self._ns.pop()
            assert x == len(self._ns)


    # def push_scope(self):
    #     self.ns.append({})

    # def pop_scope(self):
    #     return self.ns.pop()

    def __len__(self):
        return len(self._ns)

    def get_all_vars(self):
        d = {}
        for scope in reversed(self._ns):
            d.update(scope) # Overwrite upper scopes
        return d

    def get_all_vars_with_rank(self):
        d = {}
        for i, scope in enumerate(reversed(self._ns)):
            for k, v in scope.items():
                if k not in d:
                    d[k] = i, v
        return d


class AccessLevels:
    COMPILE = 1
    # COMPILE_TEXT
    QUERY = 2
    EVALUATE = 3
    READ_DB = 4
    WRITE_DB = 5


class State:

    def __init__(self, interp, db, display, ns=None):
        self.db = db
        self.interp = interp
        self.display = display
        # Add logger?

        self.tick = [0]

        self._cache = {}


    def connect(self, uri, auto_create=False):
        from preql.sql_interface import ConnectError, create_engine
        logger.info(f"[Preql] Connecting to {uri}")
        try:
            self.db = create_engine(uri, self.db._print_sql, auto_create)
        except NotImplementedError as e:
            raise Signal.make(T.NotImplementedError, None, *e.args) from e
        except ConnectError as e:
            raise Signal.make(T.DbConnectionError, None, *e.args) from e
        except ValueError as e:
            raise Signal.make(T.ValueError, None, *e.args) from e

        self._db_uri = uri


    def unique_name(self, obj):
        self.tick[0] += 1
        return obj + str(self.tick[0])


class ThreadState:
    def __init__(self, state, ns=None):
        self.state = state
        # Add logger?

        self.ns = Namespace(ns)
        self.access_level = AccessLevels.WRITE_DB
        self.stacktrace = []

    @classmethod
    def from_components(cls, interp, db, display, ns=None):
        state = State(interp, db, display)
        return cls(state, ns)

    @property
    def interp(self):
        return self.state.interp

    @property
    def db(self):
        return self.state.db

    @property
    def display(self):
        return self.state.display
    

    @classmethod
    def clone(cls, inst):
        s = cls(inst.state)
        s.ns = copy(inst.ns)
        s.access_level = inst.access_level
        s.stacktrace = copy(inst.stacktrace)
        return s


    def limit_access(self, new_level):
        return self.reduce_access(min(new_level, self.access_level))

    def reduce_access(self, new_level):
        assert new_level <= self.access_level
        s = copy(self)
        s.access_level = new_level
        return s

    def require_access(self, level):
        if self.access_level < level:
            raise InsufficientAccessLevel(level)
    def catch_access(self, level):
        if self.access_level < level:
            raise Exception("Bad access. Security risk.")


    def get_all_vars(self):
        return self.ns.get_all_vars()

    def get_all_vars_with_rank(self):
        return self.ns.get_all_vars_with_rank()

    def get_var(self, name):
        try:
            return self.ns.get_var(name)
        except NameNotFound:
            builtins = self.ns.get_var('__builtins__')
            assert builtins.type <= T.module
            try:
                return builtins.namespace[name]
            except KeyError:
                pass

            raise Signal.make(T.NameError, name, f"Name '{name}' is not defined")


    def set_var(self, name, value):
        return self.ns.set_var(name, value)

    def use_scope(self, scope: dict):
        return self.ns.use_scope(scope)

    def __copy__(self):
        return self.clone(self)

    def connect(self, uri, auto_create=False):
        return self.state.connect(uri, auto_create)

    def unique_name(self, obj):
        return self.state.unique_name(obj)



def set_var(name, value):
    return context.state.set_var(name, value)

def use_scope(scope):
    return context.state.use_scope(scope)

def get_var(name):
    return context.state.get_var(name)

def get_db_target():
    return context.state.db.target

def get_db():
    return context.state.db

def unique_name(prefix):
    return context.state.unique_name(prefix)

def require_access(access):
	return context.state.require_access(access)

def catch_access(access):
	return context.state.catch_access(access)

def get_access_level():
	return context.state.access_level

def reduce_access(new_level):
	return context.state.reduce_access(new_level)

def get_display():
	return context.state.display


from logging import getLogger, Formatter, StreamHandler, basicConfig
from logging import DEBUG, INFO, WARN, ERROR, CRITICAL

basicConfig(level=INFO, format="(%(levelname)s) %(name)s -- %(message)s")
#             )#datefmt='%m-%d %H:%M')

sh = StreamHandler()
sh.setFormatter(Formatter('%(message)s'))

def make_logger(name, level):
    logger = getLogger(name)
    logger.propagate = False
    logger.setLevel(level)
    logger.addHandler(sh)
    return logger



sql_log = make_logger('sql_output', DEBUG)
ac_log = make_logger('autocomplete', CRITICAL)
repl_log = make_logger('repl', INFO)
test_log = make_logger('tests', ERROR)



from contextlib import suppress
from typing import Union
from datetime import datetime
from dataclasses import field
from decimal import Decimal
from collections import defaultdict, deque

import arrow
import runtype
from runtype.typesystem import TypeSystem

from preql.utils import dataclass
from .base import Object

global_methods = {}

class Id:
    def __init__(self, *parts):
        assert all(isinstance(p, str) for p in parts), parts
        self.parts = parts
    def __repr__(self):
        return 'Id(%s)' % '.'.join(self.parts)
    def __str__(self):
        raise Exception("Operation not allowed!")
    def __hash__(self):
        raise Exception("Operation not allowed!")
    @property
    def repr_name(self):
        return self.parts[-1]


def _repr_type_elem(t, depth):
    return _repr_type(t, depth-1) if isinstance(t, Type) else repr(t)

def _repr_type(t, depth=2):
    if t.elems:
        if depth > 0:
            if isinstance(t.elems, dict):
                elems = '[%s]' % ', '.join(f'{k}: {_repr_type_elem(v, depth)}' for k,v in t.elems.items())
            else:
                elems = '[%s]' % ', '.join(_repr_type_elem(e, depth) for e in t.elems)
        else:
            elems = '[...]'
    else:
        elems = ''
    return f'{t._typename_with_q}{elems}'

ITEM_NAME = 'item'

@dataclass
class Type(Object):
    typename: str
    supertypes: frozenset
    elems: Union[tuple, dict] = field(hash=False, default_factory=dict)
    options: dict = field(hash=False, compare=False, default_factory=dict)
    proto_attrs: dict = field(hash=False, compare=False, default_factory=lambda: dict(global_methods))
    _nullable: bool = field(default_factory=bool)

    @property
    def _typename_with_q(self):
        n = '?' if self._nullable else ''
        return f'{self.typename}{n}'

    @property
    def elem(self):
        if isinstance(self.elems, dict):
            elem ,= self.elems.values()
        else:
            elem ,= self.elems
        return elem

    def as_nullable(self):
        # assert not self.maybe_null()
        return self.replace(_nullable=True)

    def maybe_null(self):
        return self._nullable or self is T.nulltype

    def supertype_chain(self):
        res = {
            t2
            for t1 in self.supertypes
            for t2 in t1.supertype_chain()
        }

        assert self not in res
        return res | {self}

    def __eq__(self, other, memo=None):
        "Repetitive nested equalities are assumed to be true"


        if not isinstance(other, Type):
            return False

        if memo is None:
            memo = set()

        a, b = id(self), id(other)
        if (a,b) in memo or (b,a) in memo:
            return True

        memo.add((a, b))

        l1 = self.elems if isinstance(self.elems, dict) else dict(enumerate(self.elems))
        l2 = other.elems if isinstance(other.elems, dict) else dict(enumerate(other.elems))
        if len(l1) != len(l2):
            return False

        res = self.typename == other.typename and all(
            k1==k2 and v1.__eq__(v2, memo)
            for (k1,v1), (k2,v2) in zip(l1.items(), l2.items())
        )
        return res


    @property
    def elem_types(self):
        if isinstance(self.elems, dict):
            return self.elems.values()
        return self.elems

    def issubtype(self, t):
        assert isinstance(t, Type), t
        if t.typename == 'union':   # XXX a little hacky. Change to issupertype?
            return any(self.issubtype(t2) for t2 in t.elem_types)

        if self is T.nulltype:
            if t.maybe_null():
                return True

        # TODO zip should be aware of lengths
        if t.typename in (s.typename for s in self.supertype_chain()):
            return all(e1.issubtype(e2) for e1, e2 in zip(self.elem_types, t.elem_types))
        return False

    def __le__(self, other):
        return self.issubtype(other)

    def __getitem__(self, elems):
        # TODO assert elems = (any,)
        assert not isinstance(elems, tuple), (self, elems)
        elems = {ITEM_NAME: elems}
        return self.replace(elems=elems)

    def __call__(self, elems=None, **options):
        return self.replace(elems=elems or self.elems, proto_attrs=dict(self.proto_attrs), options={**self.options, **options})

    def __repr__(self):
        # TODO Move to dp_inst?
        return _repr_type(self)

    def get_attr(self, attr):
        if self is T.unknown:
            return self

        if isinstance(self.elems, dict):
            with suppress(KeyError):
                return self.elems[attr]

        with suppress(KeyError):
            return self.proto_attrs[attr]

        return super().get_attr(attr)

    def all_attrs(self):
        # return {'elems': self.elems}
        if isinstance(self.elems, dict):
            return self.elems
        return {}

    def repr(self):
        return repr(self)

    def __or__(self, other):
        return T.union[self, other]

class TupleType(Type):
    def __getitem__(self, elems):
        assert not self.elems
        return self.replace(elems=tuple(elems))

    def __or__(self, other):
        return self.replace(elems=self.elems + (other,))

class SumType(TupleType):
    def issubtype(self, other):
        return all(t.issubtype(other) for t in self.elem_types)

class ProductType(TupleType):
    def issubtype(self, other):
        return all(a.issubtype(b) for a, b in zip(self.elem_types, other.elem_types))

class PhantomType(Type):
    def issubtype(self, other):
        return super().issubtype(other) or self.elem.issubtype(other)


class TypeDict(dict):

    def _register(self, name, supertypes=(), elems=(), type_class=Type):
        t = type_class(name, frozenset(supertypes), elems)
        assert name not in self
        T[name] = t
        dict.__setattr__(self, name, t)

    def __setattr__(self, name, args):
        if isinstance(args, tuple):
            self._register(name, *args)
        else:
            self._register(name, args)



T = TypeDict()

T.any = ()
T.unknown = [T.any]

# T.union = [T.any]
T._register('union', type_class=SumType)

T.type = [T.any]
Type.type = T.type

T.object = [T.any]
T.nulltype = [T.object]

T.primitive = [T.object]

T.text = [T.primitive]
T._rich = [T.text]
T.string = [T.text]
T.number = [T.primitive]
T.int = [T.number]
T.float = [T.number]
T.bool = [T.primitive]    # number?
T.decimal = [T.number]

T.datetime = [T.primitive]    # struct?

T.container = [T.object]

T.struct = [T.container]
T.row = [T.struct]

# T.collection = [T.container], {}
# T.table = [T.container], {}
T._register('table', [T.container], {})

T.list = [T.table], {ITEM_NAME: T.any}
T.set = [T.table], {ITEM_NAME: T.any}
T.t_id = [T.primitive], (T.table,)
T.t_relation = [T.number], (T.any,)   # t_id?

# XXX sequence instead of container?
T._register('aggregated', [T.container], (T.any,), type_class=PhantomType)
T._register('projected', [T.container], (T.any,), type_class=PhantomType)

T.json = [T.container], (T.any,)
T.json_array = [T.json]

T._register('function', [T.object], type_class=TupleType)
T.property = [T.object]

T.module = [T.object]

T.signal = [T.object]
#-----------

T.Exception = [T.signal]

T.IOError = [T.Exception]
T.CodeError = [T.Exception]
T.EvalError = [T.Exception]

# CodeError - Failures due to inherently unexecutable code
T.SyntaxError = [T.CodeError]
T.NotImplementedError = [T.CodeError]

# IOError - All errors resulting directly from attempts at I/O communication
T.FileError = [T.IOError]
T.DbError = [T.IOError]
T.DbQueryError = [T.DbError]
T.DbConnectionError = [T.DbError]

# EvalError - Errors that arise only when evaluating the code (either at run-time or compile-time)
T.TypeError = [T.EvalError]
T.ValueError = [T.EvalError]
T.NameError = [T.EvalError]
T.JoinError = [T.EvalError]
T.CompileError = [T.EvalError]

T.AttributeError = [T.NameError]
T.AssertError = [T.ValueError]
T.IndexError = [T.ValueError]
T.CastError = [T.TypeError]

T.ImportError = [T.Exception]


def _get_subtypes():
    d = defaultdict(list)
    for t in T.values():
        for st in t.supertypes:
            d[st].append(t)
    return dict(d)

subtypes = _get_subtypes()

#-------------


_t = {
    bool: T.bool,
    int: T.int,
    float: T.float,
    str: T.string,
    datetime: T.datetime,
    Decimal: T.decimal,
    arrow.Arrow: T.datetime,
}
def from_python(t):
    # TODO throw proper exception if this fails
    return _t[t]


def common_type(t1, t2):
    "Returns a type which is the closest ancestor of both t1 and t2"
    v1 = {t1}
    v2 = {t2}

    o1 = deque([t1])
    o2 = deque([t2])
    while o1 or o2:
        x1 = o1.popleft()
        v1.add(x1)
        if x1 in v2:
            return x1
        o1 += [t for t in x1.supertypes if t not in v1]

        x2 = o2.popleft()
        v2.add(x2)
        if x2 in v1:
            return x2
        o2 += [t for t in x2.supertypes if t not in v2]

    assert False


def union_types(types):
    # TODO flatten unions, remove duplications and subtypes
    ts = set(types)
    if len(ts) > 1:
        elem_type = T.union(elems=tuple(ts))
    else:
        elem_type ,= ts
    return elem_type


class ProtoTS(TypeSystem):
    def issubclass(self, t1, t2):
        if t2 is object:
            return True
        is_t2 = isinstance(t2, Type)
        if isinstance(t1, Type):
            return is_t2 and t1 <= t2
        elif is_t2:
            return False

        # Regular Python types
        return runtype.issubclass(t1, t2)

    default_type = object


class TS_Preql(ProtoTS):
    def get_type(self, obj):
        try:
            return obj.type
        except AttributeError:
            return type(obj)


class TS_Preql_subclass(ProtoTS):
    def get_type(self, obj):
        # Preql objects
        if isinstance(obj, Type):
            return obj

        # Regular Python
        return type(obj)


dp_type = runtype.Dispatch(TS_Preql_subclass())
dp_inst = runtype.Dispatch(TS_Preql())
